/*
 *	Copyright (c) 2010, University of Helsinki
 *	
 *	All rights reserved.
 *	
 *	Redistribution and use in source and binary forms, with
 *	or without modification, are permitted provided that the
 *	following conditions are met:
 *	
 *		1. Redistributions of source code must retain the
 *		   above copyright notice, this list of conditions
 *		   and the following disclaimer.
 *	
 *		2. Redistributions in binary form must reproduce the
 *		   above copyright notice, this list of conditions and
 *		   the following disclaimer in the documentation and/or
 *		   other materials provided with the distribution.
 *	
 *		3. Neither the name of the University of Helsinki
 *		   nor the names of its contributors may be used
 *		   to endorse or promote products derived from this
 *		   software without specific prior written permission.
 *	
 *	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
 *	CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
 *	INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *	OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *	ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
 *	CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *	SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *	NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *	LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *	HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *	CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 *	OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 *	EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <string.h>
#include "types.h"
#include "csum.h"

int
icmpgen(int type, int code, int u1, int u2, int u3, int u4, u32int *origin, uchar *buf, int pktlen, int buflen)
{
	int snaplen;
	u16int csum;
	uchar dst[4];
	uchar src[4];

	memmove(dst, buf + 16, 4);
	memmove(src, buf + 12, 4);

	snaplen = pktlen;
	if(snaplen > buflen - 28)
		snaplen = buflen - 28;

	memmove(buf + 28, buf, snaplen);	/* save tcp data */

/* build ip header */

	buf[0] = 0x45;			/* version 4, len=5*4 bytes */
	buf[1] = 0;			/* TOS */
	buf[2] = 0;			/* len hi */
	buf[3] = 28 + snaplen;		/* len lo */
	buf[4] = 0;			/* ip id hi */
	buf[5] = 0;			/* ip id lo */
	buf[6] = 0;			/* fragoff hi */
	buf[7] = 0;			/* fragoff lo */
	buf[8] = 255;			/* ttl */
	buf[9] = 1;			/* protocol: ICMP */
	buf[10] = 0;			/* ip checksum hi */
	buf[11] = 0;			/* ip checksum lo */
	memmove(buf + 12, origin, 4);	/* src address */
	memmove(buf + 16, src, 4);	/* dst address */

/* buf[20..] possible ip options */

	csum = ipsum(buf, 20);
	buf[10] = csum & 0xff;
	buf[11] = csum >> 8;

/* build icmp header */

	buf[20] = type;			/* type: explicit loss */
	buf[21] = code;			/* code: whatever */

	buf[22] = 0;			/* icmp checksum hi */
	buf[23] = 0;			/* icmp checksum lo */

	buf[24] = u1;			/* MBZ */
	buf[25] = u2;
	buf[26] = u3;
	buf[27] = u4;

	csum = ipsum(buf+20, 8 + snaplen);
	buf[22] = csum & 0xff;	/* checksum lo */
	buf[23] = csum >> 8;	/* checksum hi */

	return 28 + snaplen;
}
