/*
 *	Copyright (c) 2010, University of Helsinki
 *	
 *	All rights reserved.
 *	
 *	Redistribution and use in source and binary forms, with
 *	or without modification, are permitted provided that the
 *	following conditions are met:
 *	
 *		1. Redistributions of source code must retain the
 *		   above copyright notice, this list of conditions
 *		   and the following disclaimer.
 *	
 *		2. Redistributions in binary form must reproduce the
 *		   above copyright notice, this list of conditions and
 *		   the following disclaimer in the documentation and/or
 *		   other materials provided with the distribution.
 *	
 *		3. Neither the name of the University of Helsinki
 *		   nor the names of its contributors may be used
 *		   to endorse or promote products derived from this
 *		   software without specific prior written permission.
 *	
 *	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
 *	CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
 *	INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *	OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *	ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
 *	CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *	SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *	NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *	LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *	HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *	CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 *	OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 *	EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <fcntl.h>
#include <net/route.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <linux/if.h>
#include <linux/if_tun.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>

#include <netinet/in.h>
#include <arpa/inet.h>

#include "error.h"

void
ifset(int fd, struct ifreq *ifr, char *ip, int flags)
{
	struct sockaddr_in *ifaddr;

	ifaddr = (struct sockaddr_in *)&ifr->ifr_addr;

	ifaddr->sin_family = AF_INET;
	if(inet_pton(AF_INET, ip, &ifaddr->sin_addr) == -1)
		error("inet_pton %s", ip);
	if(ioctl(fd, flags, ifr) == -1)
		error("ioctl");
}

void
ifsetqlen(int fd, struct ifreq *ifr, int val)
{
	ifr->ifr_qlen = val;
	if(ioctl(fd, SIOCSIFTXQLEN, ifr) == -1)
		error("ioctl");
}

int
opentun(char *locip, char *remip)
{
	struct ifreq ifr;
	int fd, Tfd;

	if((fd = open("/dev/net/tun", O_RDWR)) < 0)
		error("open /dev/net/tun");

	/* no packet info and no ethernet crap, please. */
	memset(&ifr, 0, sizeof(ifr));
	ifr.ifr_flags = IFF_TUN | IFF_NO_PI;
	if(ioctl(fd, TUNSETIFF, (void *) &ifr) == -1)
		error("ioctl");

	Tfd = socket(AF_INET, SOCK_DGRAM, 0);

	ifset(Tfd, &ifr, locip, SIOCSIFADDR);
	ifset(Tfd, &ifr, remip, SIOCSIFDSTADDR);
	ifset(Tfd, &ifr, "255.255.255.255", SIOCSIFNETMASK);
	ifsetqlen(Tfd, &ifr, 100);

	ifr.ifr_flags |= (IFF_UP | IFF_POINTOPOINT);
	if(ioctl(Tfd, SIOCSIFFLAGS, &ifr) == -1)
		error("ioctl");

	close(Tfd);

	return fd;
}
