/*
 *	Copyright (c) 2010, University of Helsinki
 *	
 *	All rights reserved.
 *	
 *	Redistribution and use in source and binary forms, with
 *	or without modification, are permitted provided that the
 *	following conditions are met:
 *	
 *		1. Redistributions of source code must retain the
 *		   above copyright notice, this list of conditions
 *		   and the following disclaimer.
 *	
 *		2. Redistributions in binary form must reproduce the
 *		   above copyright notice, this list of conditions and
 *		   the following disclaimer in the documentation and/or
 *		   other materials provided with the distribution.
 *	
 *		3. Neither the name of the University of Helsinki
 *		   nor the names of its contributors may be used
 *		   to endorse or promote products derived from this
 *		   software without specific prior written permission.
 *	
 *	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
 *	CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
 *	INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *	OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *	ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
 *	CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *	SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *	NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *	LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *	HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *	CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 *	OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 *	EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdarg.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>

int dflag;

static int
vreport(char *type, char *fmt, va_list vp)
{
	int oerrno;
	int n;
	oerrno = errno;
	n = fprintf(stderr, "%s: ", type);
	n += vfprintf(stderr, fmt, vp);
	if(oerrno != 0)
		n += fprintf(stderr, ": %s", strerror(errno));
	n += fprintf(stderr, "\n");
	return n;
}

int
error(char *fmt, ...)
{
	va_list vp;
	va_start(vp, fmt);
	vreport("error", fmt, vp);
	va_end(vp);
	exit(1);
}

int
warn(char *fmt, ...)
{
	va_list vp;
	int r;
	va_start(vp, fmt);
	r = vreport("warn", fmt, vp);
	va_end(vp);
	errno = 0;
	return r;
}

int
info(char *fmt, ...)
{
	va_list vp;
	int r;
	if(!dflag)
		return 0;
	va_start(vp, fmt);
	r = vreport("info", fmt, vp);
	va_end(vp);
	errno = 0;
	return r;
}
