/*
 *	Copyright (c) 2010, University of Helsinki
 *	
 *	All rights reserved.
 *	
 *	Redistribution and use in source and binary forms, with
 *	or without modification, are permitted provided that the
 *	following conditions are met:
 *	
 *		1. Redistributions of source code must retain the
 *		   above copyright notice, this list of conditions
 *		   and the following disclaimer.
 *	
 *		2. Redistributions in binary form must reproduce the
 *		   above copyright notice, this list of conditions and
 *		   the following disclaimer in the documentation and/or
 *		   other materials provided with the distribution.
 *	
 *		3. Neither the name of the University of Helsinki
 *		   nor the names of its contributors may be used
 *		   to endorse or promote products derived from this
 *		   software without specific prior written permission.
 *	
 *	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
 *	CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
 *	INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *	OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *	ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
 *	CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *	SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *	NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *	LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *	HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *	CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 *	OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 *	EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <sys/time.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <time.h>
#include <fcntl.h>
#include <poll.h>

#include <errno.h>
#include <string.h>

#include "error.h"
#include "types.h"

int
read_to(int fd, void *vp, int len, int to)
{
	struct pollfd pfd;
	pfd.fd = fd;
	pfd.events = POLLIN;
	switch(poll(&pfd, 1, to)){
	case -1:
		error("bad select");
	case 0:
		return -1;
	default:
		return read(fd, vp, len);
	}
}

int
main(int argc, char *argv[])
{
	int i, j;
	int opt;
	char *p;
	int *fdtab, fd;
	struct sockaddr_in sa;
	char buf[64];
	int port;
	int maxfds = 8192;
	char *ipaddr;
	char reason[64];

	while((opt = getopt(argc, argv, "m:")) != -1)
	switch(opt){
	case 'm':
		maxfds = strtol(optarg, &p, 10);
		if(*p != '\0')
			goto usage;
		break;
	default:
	usage:
		printf("usage: %s [-m maxcon] srv-ip4-addr port\n", argv[0]);
		exit(1);
	}

	setbuf(stdout, NULL);

	if(optind != argc - 2)
		goto usage;

	ipaddr = argv[optind++];
	port = strtol(argv[optind++], NULL, 10);

	sa.sin_family = AF_INET;
	sa.sin_addr.s_addr = inet_addr(ipaddr);
	sa.sin_port = htons(port);

	fdtab = malloc(sizeof fdtab[0] * maxfds);

	reason[0] = '\0';
	for(i = 0; i < maxfds; i++){

printf("ncon %d -- ", i);
		if((fdtab[i] = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)) == -1)
			error("socket");

		if(connect(fdtab[i], (struct sockaddr*)&sa, sizeof sa) == -1){
			snprintf(reason, sizeof reason, "connect %d: %s", i, strerror(errno));
			goto done;
		}

		for(j = 0; j < i; j++){
			if(write(fdtab[j], "H", 1) == -1){
				snprintf(reason, sizeof reason, "write %d: %s", i, strerror(errno));
				goto done;
			}
		}
		for(j = 0; j < i; j++){
			if(read_to(fdtab[j], buf, sizeof buf, 10000) == -1){
				snprintf(reason, sizeof reason, "read %d: %s", i, strerror(errno));
				goto done;
			}
		}
printf("ok\n");
	}

done:
	for(j = 0; j <= i; j++)
		close(fdtab[j]);

	fd = creat("data", 0666);
	snprintf(buf, sizeof buf, "%s %d 31 %d #reason:%s\n",
		inet_ntoa(sa.sin_addr), port,
		i, reason);
	write(fd, buf, strlen(buf));
	close(fd);

	return 0;
}

