/*
 *	Copyright (c) 2010, University of Helsinki
 *	
 *	All rights reserved.
 *	
 *	Redistribution and use in source and binary forms, with
 *	or without modification, are permitted provided that the
 *	following conditions are met:
 *	
 *		1. Redistributions of source code must retain the
 *		   above copyright notice, this list of conditions
 *		   and the following disclaimer.
 *	
 *		2. Redistributions in binary form must reproduce the
 *		   above copyright notice, this list of conditions and
 *		   the following disclaimer in the documentation and/or
 *		   other materials provided with the distribution.
 *	
 *		3. Neither the name of the University of Helsinki
 *		   nor the names of its contributors may be used
 *		   to endorse or promote products derived from this
 *		   software without specific prior written permission.
 *	
 *	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
 *	CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
 *	INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *	OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *	ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
 *	CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *	SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *	NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *	LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *	HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *	CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 *	OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 *	EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/select.h>

#include <errno.h>

#include <signal.h>
#include <sys/wait.h>

#include <poll.h>

#include "error.h"
#include "readn.h"


unsigned char buf[512];
struct pollfd *polltab;
int *fdtab;
int nfds;

int
main(int argc, char *argv[])
{
	int lfd, opt;
	struct sockaddr_in sa;
	socklen_t len;
	char *ipaddr;
	int port;
	int npoll;

	setbuf(stdout, NULL);

	if(argc != 3){
		printf("usage: %s srv-ip4-addr port\n", argv[0]);
		exit(1);
	}

	ipaddr = argv[1];
	port = strtol(argv[2], NULL, 10);

	setpgid(0,0);

	if((lfd = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)) == -1)
		error("socket");

	opt = 1;
	setsockopt(lfd, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof opt);
	sa.sin_family = AF_INET;
	sa.sin_addr.s_addr = inet_addr(ipaddr);
	sa.sin_port = htons(port);
	if(bind(lfd, (struct sockaddr*)&sa, sizeof sa) == -1)
		error("bind");
	if(listen(lfd, 1) == -1)
		error("listen");

	nfds = 1;
	polltab = malloc(sizeof polltab[0] * nfds);
	fdtab = malloc(sizeof fdtab[0] * nfds);
	polltab[0].fd = lfd;
	polltab[0].events = POLLIN;
	npoll = 1;

	for(;;){
		int i, j, n;
		poll(polltab, npoll, -1);
		if(polltab[0].revents & POLLIN){
			int nfd;
			len = sizeof sa;
			if((nfd = accept(polltab[0].fd, (struct sockaddr*)&sa, &len)) == -1)
				error("accept\n");
			if(npoll == nfds){
				nfds *= 2;
				polltab = realloc(polltab, sizeof polltab[0] * nfds);
			}
			polltab[npoll].fd = nfd;
			polltab[npoll].events = POLLIN;
			polltab[npoll].revents = 0;
			npoll++;
		}
		for(i = 1; i < npoll; i++){
		skippost:
			if(polltab[i].revents & POLLIN){
				if((n = read(polltab[i].fd, buf, sizeof buf)) <= 0){
				caseclosed:
printf("closing %d\n", i);
					close(polltab[i].fd);
					npoll--;
					memmove(polltab + i, polltab + i + 1, sizeof polltab[0] * (npoll - i));
					if(i < npoll)
						goto skippost;
					continue;
				}
				for(j = 0; j < n; j++)
					if(buf[j] == 255)
						exit(0);
				if(write(polltab[i].fd, buf, n) != n)
					goto caseclosed;
			}
		}
	}
	return 0;
}

