/*
 *	Copyright (c) 2010, University of Helsinki
 *	
 *	All rights reserved.
 *	
 *	Redistribution and use in source and binary forms, with
 *	or without modification, are permitted provided that the
 *	following conditions are met:
 *	
 *		1. Redistributions of source code must retain the
 *		   above copyright notice, this list of conditions
 *		   and the following disclaimer.
 *	
 *		2. Redistributions in binary form must reproduce the
 *		   above copyright notice, this list of conditions and
 *		   the following disclaimer in the documentation and/or
 *		   other materials provided with the distribution.
 *	
 *		3. Neither the name of the University of Helsinki
 *		   nor the names of its contributors may be used
 *		   to endorse or promote products derived from this
 *		   software without specific prior written permission.
 *	
 *	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
 *	CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
 *	INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *	OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *	ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
 *	CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *	SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *	NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *	LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *	HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *	CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 *	OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 *	EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <sys/time.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <time.h>
#include <fcntl.h>

#include "error.h"
#include "types.h"

int concur = 80;

u64int
time_ms(void)
{
	u64int ms;
	struct timeval tv;
	gettimeofday(&tv, NULL);
	ms = tv.tv_sec * 1000;
	ms += tv.tv_usec / 1000;
	return ms;
}

void
sigalrm(int sig)
{
	sig=sig;
}

int
read_to(int fd, void *vp, int len, int to)
{
	fd_set rset;
	struct timeval tv;
	FD_ZERO(&rset);
	FD_SET(fd, &rset);
	if(to <= 0)
		to = 1;
	tv.tv_sec = to / 1000;
	tv.tv_usec = (to % 1000) * 1000;
	switch(select(fd+1, &rset, NULL, NULL, &tv)){
	case -1:
		error("bad select");
	case 0:
		return -1;
	default:
		return read(fd, vp, len);
	}
}

int
main(int argc, char *argv[])
{
	int i, n, nto;
	int *fd;
	struct sockaddr_in sa;
	char buf[64];
	struct sigaction act;
	int lo_ms, hi_ms, result_ms, cum_sleep, *sleep_ms;
	u64int *start_ms;
	int step, spread, slack;
	int port;
	char *ipaddr;

	setbuf(stdout, NULL);

	if(argc != 3){
		printf("usage: %s srv-ip4-addr port\n", argv[0]);
		exit(1);
	}

	/* set sighand so read() gets interrupted by SIGALRM */
	memset(&act, 0, sizeof act);	
	act.sa_handler = sigalrm;
	sigfillset(&act.sa_mask);
	sigaction(SIGALRM, &act, NULL);

	lo_ms = 0;
	hi_ms = 24*60*60*1000;
	//hi_ms = 60*60*1000;
	sleep_ms = malloc(concur * sizeof sleep_ms[0]);
	fd = malloc(concur * sizeof fd[0]);
	start_ms = malloc(concur * sizeof start_ms[0]);

	ipaddr = argv[1];
	port = strtol(argv[2], NULL, 10);

	sa.sin_family = AF_INET;
	sa.sin_addr.s_addr = inet_addr(ipaddr);
	sa.sin_port = htons(port);

	while(hi_ms - lo_ms > 200) {

			step = (hi_ms - lo_ms) / (concur-1);
			spread = concur * step;
			slack = (spread - (hi_ms - lo_ms)) / 2;
printf("lo_ms:%d hi_ms:%d step:%d spread:%d slack:%d\n", lo_ms, hi_ms, step, spread, slack);
	
		for(i = 0; i < concur; i++){
			if((fd[i] = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)) == -1)
				error("socket");
			if(connect(fd[i], (struct sockaddr*)&sa, sizeof sa) == -1)
				error("connect");

			sleep_ms[i] = i * step + lo_ms - slack;
			n = snprintf(buf + 2, sizeof buf - 2, "%d", sleep_ms[i]);
			buf[0] = 0;
			buf[1] = n;
			if(write(fd[i], buf, n + 2) != n + 2)
				error("write");
			start_ms[i] = time_ms();
		}

		cum_sleep = 0;
		nto = 0;
		for(i = 0; i < concur; i++){
			int timeout = start_ms[i] + sleep_ms[i] + 1000 - time_ms();
			if((n = read_to(fd[i], buf, sizeof buf, timeout)) > 0){
				nto = 0;
				printf("through %d\n", sleep_ms[i]);
				if(lo_ms < sleep_ms[i])
					lo_ms = sleep_ms[i];
			} else {
				nto++;
				printf("timeout %d\n", sleep_ms[i]);
				if(nto > 1 && hi_ms > sleep_ms[i-1]){
					hi_ms = sleep_ms[i-1];
					while(i < concur){
						close(fd[i]);
						i++;
					}
					break;
				}
			}
			cum_sleep = sleep_ms[i];
			close(fd[i]);
		}
	}

	if((fd[0] = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)) == -1)
		error("socket");
	if(connect(fd[0], (struct sockaddr*)&sa, sizeof sa) == -1)
		error("connect");
	buf[0] = 250;
	write(fd[0], buf, 1);
	close(fd[0]);


printf("lo_ms:%d hi_ms:%d\n", lo_ms, hi_ms);

	result_ms = (hi_ms + lo_ms) / 2;

printf("result %d.%03d\n", result_ms / 1000, result_ms % 1000);

	fd[0] = creat("data", 0666);
	snprintf(buf, sizeof buf, "%s %d 31 %d.%03d\n",
		inet_ntoa(sa.sin_addr), port,
		result_ms / 1000, result_ms % 1000);
	write(fd[0], buf, strlen(buf));
	close(fd[0]);

	return 0;
}

