/*
 *	Copyright (c) 2010, University of Helsinki
 *	
 *	All rights reserved.
 *	
 *	Redistribution and use in source and binary forms, with
 *	or without modification, are permitted provided that the
 *	following conditions are met:
 *	
 *		1. Redistributions of source code must retain the
 *		   above copyright notice, this list of conditions
 *		   and the following disclaimer.
 *	
 *		2. Redistributions in binary form must reproduce the
 *		   above copyright notice, this list of conditions and
 *		   the following disclaimer in the documentation and/or
 *		   other materials provided with the distribution.
 *	
 *		3. Neither the name of the University of Helsinki
 *		   nor the names of its contributors may be used
 *		   to endorse or promote products derived from this
 *		   software without specific prior written permission.
 *	
 *	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
 *	CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
 *	INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *	OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *	ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
 *	CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *	SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *	NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *	LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *	HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *	CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 *	OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 *	EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/select.h>

#include <errno.h>

#include <signal.h>
#include <sys/wait.h>

#include "error.h"
#include "readn.h"


unsigned char buf[258];

void
msleep(int ms)
{
	struct timeval tv;
	tv.tv_sec = ms / 1000;
	tv.tv_usec = (ms % 1000) * 1000;
	select(0, NULL, NULL, NULL, &tv);
}

void
sigchild(int sig)
{
	int oerrno;
	oerrno = errno;
	while(waitpid(-1, NULL, WNOHANG) > 0)
		;
	errno = oerrno;
}

int
main(int argc, char *argv[])
{
	int fd, lfd, opt;
	int sleep_ms;
	struct sockaddr_in sa;
	struct sigaction act, oact;
	socklen_t len;
	char *ipaddr;
	int port;

	setbuf(stdout, NULL);

	if(argc != 3){
		printf("usage: %s srv-ip4-addr port\n", argv[0]);
		exit(1);
	}

	ipaddr = argv[1];
	port = strtol(argv[2], NULL, 10);

	setpgid(0,0);

	if((lfd = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)) == -1)
		error("socket");

	opt = 1;
	setsockopt(lfd, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof opt);
	sa.sin_family = AF_INET;
	sa.sin_addr.s_addr = inet_addr(ipaddr);
	sa.sin_port = htons(port);
	if(bind(lfd, (struct sockaddr*)&sa, sizeof sa) == -1)
		error("bind");
	if(listen(lfd, 1) == -1)
		error("listen");

	memset(&act, 0, sizeof act);
	act.sa_handler = sigchild;
	act.sa_flags = SA_RESTART;
	sigfillset(&act.sa_mask);
	sigaction(SIGCHLD, &act, &oact);

	for(;;){
		len = sizeof sa;
		if((fd = accept(lfd, (struct sockaddr*)&sa, &len)) == -1){
			if(errno == EINTR)
				continue;
			printf("accept %s\n", strerror(errno));
			return 0;
		}
reread:
		if(read(fd, buf, 1) != 1){
			if(errno == EINTR)
				goto reread;
			printf("read 1 %s\n", strerror(errno));
			close(fd);
			continue;
		}
		if(buf[0] == 250)
			goto out;
		switch(fork()){
		case -1:
			error("fork");
		case 0:
			sigaction(SIGCHLD, &oact, NULL);
			break;
		default:
			close(fd);
			continue;
		}
		switch(buf[0]){
		case 250:
			goto out;
		case 0:
			if(read(fd, buf + 1, 1) != 1)
				error("read 1");
			if(readn(fd, buf + 2, buf[1]) != buf[1])
				error("read %d", buf[1]);
			buf[buf[1]+2] = 0;
			sleep_ms = strtol((char*)buf+2, NULL, 10);
			printf("sleep %d ms\n", sleep_ms);
			msleep(sleep_ms);
			buf[0] = 250;
			write(fd, buf, 1);
			close(fd);
			break;
		default:
			error("invalid command %d", buf[0]);
		}
	}
out:
	kill(0, SIGKILL);
	return 0;
}

