/*
 *	Copyright (c) 2010, University of Helsinki
 *	
 *	All rights reserved.
 *	
 *	Redistribution and use in source and binary forms, with
 *	or without modification, are permitted provided that the
 *	following conditions are met:
 *	
 *		1. Redistributions of source code must retain the
 *		   above copyright notice, this list of conditions
 *		   and the following disclaimer.
 *	
 *		2. Redistributions in binary form must reproduce the
 *		   above copyright notice, this list of conditions and
 *		   the following disclaimer in the documentation and/or
 *		   other materials provided with the distribution.
 *	
 *		3. Neither the name of the University of Helsinki
 *		   nor the names of its contributors may be used
 *		   to endorse or promote products derived from this
 *		   software without specific prior written permission.
 *	
 *	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
 *	CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
 *	INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *	OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *	ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
 *	CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *	SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *	NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *	LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *	HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *	CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 *	OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 *	EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *	
 *	Contributors:
 *	
 *		Seppo Hätönen
 *		Aki Nyrhinen
 */


#include <stdlib.h> 
#include <sys/types.h>
#include <sys/socket.h>
#include <signal.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h> /* memset() */
#include <sys/time.h> /* select() */ 

#define REMOTE_SERVER_PORT 11000
#define MAX_MSG 1024

#define SOCKET_ERROR -1


char msg[MAX_MSG];
char recline[MAX_MSG+1];


void sigalrm(int sig){
	sig=sig;
}


/* Basic timeout test without ack */
int test1(int contfd, int testfd, struct sockaddr_in servAddr, socklen_t servLen, int incr){
	
	int timeout = 1;
	int flags = 0;
	int n;
	int inc = incr;

	printf("Starting timeout test with increment of %d, ack.\n", inc);
	
	/* Register test address */
	printf("Sending \"test\" to %s:%u timeout=%ds\n", inet_ntoa(servAddr.sin_addr),ntohs(servAddr.sin_port), timeout);		
	
	sprintf(msg, "test");
	n=sendto(testfd, msg, strlen(msg), flags, (struct sockaddr *)&servAddr, sizeof(servAddr));
	if(n<0){
		perror("sendto");
		return -1;		
	}
	/* Start test */
	sleep(1);
	while(1)
	{
	
	
		printf("Sending ping %d to %s:%u\n",timeout, inet_ntoa(servAddr.sin_addr),ntohs(servAddr.sin_port));		
		sprintf(msg, "ping %d",timeout);
		n=sendto(contfd, msg, strlen(msg), flags, (struct sockaddr *)&servAddr, sizeof(servAddr));
		if(n<0){
			perror("sendto");
			break;
		}
		alarm(1);
		n=recvfrom(testfd,recline,MAX_MSG,flags, (struct sockaddr *)&servAddr,&servLen);
		alarm(0);
		recline[n]=0;
		if(n>0){
			printf("Received %s from %s:%u\n",recline,inet_ntoa(servAddr.sin_addr),
					ntohs(servAddr.sin_port));
			/*
			sprintf(msg,"ack");
			n=sendto(testfd, msg, strlen(msg),flags,(struct sockaddr *)&servAddr, sizeof(servAddr));
			*/
			timeout=timeout+inc;
			
			sleep(timeout);
		}else{
			printf("timeout detected for %s at %d seconds\n",inet_ntoa(servAddr.sin_addr), timeout);
			printf("Trying retransmit\n");

			printf("Sending ping %d to %s:%u\n",timeout, inet_ntoa(servAddr.sin_addr),ntohs(servAddr.sin_port));		
			sprintf(msg, "ping %d",timeout);
			n=sendto(contfd, msg, strlen(msg), flags, (struct sockaddr *)&servAddr, sizeof(servAddr));
			if(n<0){
				perror("sendto");
				break;
			}
			alarm(1);
			
			n=recvfrom(testfd,recline,MAX_MSG,flags, (struct sockaddr *)&servAddr,&servLen);
			alarm(0);
			recline[n]=0;
			if(n>0){
				continue;
			}
			else
				break;	
		}
		if(timeout>4096){
			printf("No timeout detected within 4096 seconds\n");;
			timeout=-1;

			return -1;
		}
	}
	printf("Timeout detected between %d and %d for %s\n", timeout-inc, timeout, 
			inet_ntoa(servAddr.sin_addr));
	return timeout-inc;
}

/* Basic timeout test with ack */
int test2(int contfd, int testfd, struct sockaddr_in servAddr, socklen_t servLen, int incr){
	
	int timeout = 1;
	int flags = 0;
	int n;
	int inc = incr;

	printf("Starting timeout test with increment of %d, ack.\n", inc);
	
	/* Register test address */
	printf("Sending \"test\" to %s:%u timeout=%ds\n", inet_ntoa(servAddr.sin_addr),ntohs(servAddr.sin_port), timeout);		
	sprintf(msg, "test");
	n=sendto(testfd, msg, strlen(msg), flags, (struct sockaddr *)&servAddr, sizeof(servAddr));
	if(n<0){
		perror("sendto");
		return -1;		
	}
	/* Start test */
	sleep(1);
	while(1)
	{
	
	
		printf("Sending ping %d to %s:%u\n",timeout, inet_ntoa(servAddr.sin_addr),ntohs(servAddr.sin_port));		
		sprintf(msg, "ping %d",timeout);
		n=sendto(contfd, msg, strlen(msg), flags, (struct sockaddr *)&servAddr, sizeof(servAddr));
		if(n<0){
			perror("sendto");
			break;
		}
		alarm(1);
		n=recvfrom(testfd,recline,MAX_MSG,flags, (struct sockaddr *)&servAddr,&servLen);
		alarm(0);
		recline[n]=0;
		if(n>0){
			printf("Received %s from %s:%u\n",recline,inet_ntoa(servAddr.sin_addr),
					ntohs(servAddr.sin_port));
			sprintf(msg,"ack");
			printf("Sending ping %d to %s:%u\n",timeout, inet_ntoa(servAddr.sin_addr),ntohs(servAddr.sin_port));		
			n=sendto(testfd, msg, strlen(msg),flags,(struct sockaddr *)&servAddr, sizeof(servAddr));
			timeout=timeout+inc;
			
			sleep(timeout);
		}else{
			printf("timeout detected for %s at %d seconds\n",inet_ntoa(servAddr.sin_addr), timeout);
			printf("Trying retransmit\n");

			sprintf(msg, "ping %d",timeout);
			printf("Sending ping %d to %s:%u\n",timeout, inet_ntoa(servAddr.sin_addr),ntohs(servAddr.sin_port));		
			n=sendto(contfd, msg, strlen(msg), flags, (struct sockaddr *)&servAddr, sizeof(servAddr));
			if(n<0){
				perror("sendto");
				break;
			}
			alarm(1);
			n=recvfrom(testfd,recline,MAX_MSG,flags, (struct sockaddr *)&servAddr,&servLen);
			alarm(0);
			recline[n]=0;
			if(n>0){
				continue;
			}
			else
				break;	
		}
		if(timeout>4096){
			printf("No timeout detected within 4096 seconds\n");;
			timeout=-1;
			return -1;
		}
	}
	printf("Timeout detected between %d and %d for %s\n", timeout-inc, timeout, 
			inet_ntoa(servAddr.sin_addr));
	return timeout-inc;
}


/* Basic timeout test without ack, increase timeout fast to given amount, then start with increment */
int test3(int contfd, int testfd, struct sockaddr_in servAddr, socklen_t servLen, int incr, int skip){
	
	int timeout = 1;
	int flags = 0;
	int n;
	int inc = incr;

	printf("Starting timeout test with increment of %d, no ack.\n", inc);
	
	/* Register test address */
	sprintf(msg, "test");
	printf("Sending \"test\" %d to %s:%u\n",timeout, inet_ntoa(servAddr.sin_addr),ntohs(servAddr.sin_port));	
	n=sendto(testfd, msg, strlen(msg), flags, (struct sockaddr *)&servAddr, sizeof(servAddr));
	if(n<0){
		perror("sendto");
		exit(EXIT_FAILURE);		
	}
	/* Start test */
	sleep(1);
	while(1)
	{
	
	
		printf("Sending ping %d to %s:%u\n",timeout, inet_ntoa(servAddr.sin_addr),ntohs(servAddr.sin_port));		
		sprintf(msg, "ping %d",timeout);
		n=sendto(contfd, msg, strlen(msg), flags, (struct sockaddr *)&servAddr, sizeof(servAddr));
		if(n<0){
			perror("sendto");
			break;
		}
		alarm(1);
		n=recvfrom(testfd,recline,MAX_MSG,flags, (struct sockaddr *)&servAddr,&servLen);
		alarm(0);
		recline[n]=0;
		if(n>0){
			printf("Received %s from %s:%u\n",recline,inet_ntoa(servAddr.sin_addr),
					ntohs(servAddr.sin_port));
			/*
			sprintf(msg,"ack");
			n=sendto(testfd, msg, strlen(msg),flags,(struct sockaddr *)&servAddr, sizeof(servAddr));
			*/

			if(timeout+100<skip){
				timeout=timeout+100;
				sleep(timeout);
				continue;
			} 
		        
			if(timeout<skip){
				timeout=skip;
				sleep(timeout);
				continue;		
			}
			else{
				timeout=timeout+inc;
				sleep(timeout);
			}
		}else{
			printf("timeout detected for %s at %d seconds\n",inet_ntoa(servAddr.sin_addr), timeout);
			printf("Trying retransmit\n");

			printf("Sending ping %d to %s:%u\n",timeout, inet_ntoa(servAddr.sin_addr),ntohs(servAddr.sin_port));		
			sprintf(msg, "ping %d",timeout);
			n=sendto(contfd, msg, strlen(msg), flags, (struct sockaddr *)&servAddr, sizeof(servAddr));
			if(n<0){
				perror("sendto");
				break;
			}
			alarm(1);
			n=recvfrom(testfd,recline,MAX_MSG,flags, (struct sockaddr *)&servAddr,&servLen);
			alarm(0);
			recline[n]=0;
			if(n>0){
				continue;
			}
			else
				break;	
		}
		if(timeout>4096){
			printf("No timeout detected within 4096 seconds\n");;
			timeout=-1;
			return -1;
		}
	}
	printf("Timeout detected between %d and %d for %s\n", timeout-inc, timeout, 
			inet_ntoa(servAddr.sin_addr));
	return timeout-inc;
}



/* Basic timeout test with ack, increase timeout fast to given amount, then start with increment */
int test4(int contfd, int testfd, struct sockaddr_in servAddr, socklen_t servLen, int incr, int skip){
	
	int timeout = 0;
	int flags = 0;
	int n;
	int inc = incr;

	printf("Starting timeout test with increment of %d, no ack.\n", inc);
	
	/* Register test address */
	printf("Sending \"test\" %d to %s:%u\n",timeout, inet_ntoa(servAddr.sin_addr),ntohs(servAddr.sin_port));	
	sprintf(msg, "test");
	n=sendto(testfd, msg, strlen(msg), flags, (struct sockaddr *)&servAddr, sizeof(servAddr));
	if(n<0){
		perror("sendto");
		return -1;		
	}
	
	/* Start test */
	sleep(1);
	while(1)
	{
	
		printf("Sending ping %d to %s:%u\n",timeout, inet_ntoa(servAddr.sin_addr),ntohs(servAddr.sin_port));	
		printf("Sending ping %d\n",timeout);		
		sprintf(msg, "ping %d",timeout);
		n=sendto(contfd, msg, strlen(msg), flags, (struct sockaddr *)&servAddr, sizeof(servAddr));
		if(n<0){
			perror("sendto");
			break;
		}
		alarm(1);
		n=recvfrom(testfd,recline,MAX_MSG,flags, (struct sockaddr *)&servAddr,&servLen);
		alarm(0);
		recline[n]=0;
		if(n>0){
			printf("Received %s from %s:%u\n",recline,inet_ntoa(servAddr.sin_addr),
					ntohs(servAddr.sin_port));
			
			sprintf(msg,"ack");
			n=sendto(testfd, msg, strlen(msg),flags,(struct sockaddr *)&servAddr, sizeof(servAddr));
			
			if(timeout+100<skip){
				timeout=timeout+100;
				sleep(timeout);
				continue;
			} 
		        
			if(timeout<skip){
				timeout=skip;
				sleep(timeout);
				continue;		
			}
			else{
				timeout=timeout+inc;
				sleep(timeout);
			}
		}else{
			printf("timeout detected for %s at %d seconds\n",inet_ntoa(servAddr.sin_addr), timeout);
			printf("Trying retransmit\n");

			printf("Sending ping %d to %s:%u\n",timeout, inet_ntoa(servAddr.sin_addr),ntohs(servAddr.sin_port));		
			sprintf(msg, "ping %d",timeout);
			n=sendto(contfd, msg, strlen(msg), flags, (struct sockaddr *)&servAddr, sizeof(servAddr));
			if(n<0){
				perror("sendto");
				break;
			}
			alarm(1);
			n=recvfrom(testfd,recline,MAX_MSG,flags, (struct sockaddr *)&servAddr,&servLen);
			alarm(0);
			recline[n]=0;
			if(n>0){
				continue;
			}
			else
				break;	
		}
		if(timeout>4096){
			printf("No timeout detected within 4096 seconds\n");
			timeout=-1;
			return -1;
		}
	}
	printf("Timeout detected between %d and %d for %s\n", timeout-inc, timeout, 
			inet_ntoa(servAddr.sin_addr));
	return timeout-inc;
}

/* initial timeout, with incremental search */
int test5(int contfd, int testfd, struct sockaddr_in servAddr, socklen_t servLen, struct sockaddr_in testAddr, int testLen, int start){
	
	int timeout = start;
	int flags = 0;
	int n;
	int inc = start;

	testLen=testLen; /* shutup gcc! */
	 
	printf("Starting timeout test with increment of %d, no ack.\n", inc);
	
	/* Register test address */
	printf("Sending \"test\" %d to %s:%u\n",timeout, inet_ntoa(servAddr.sin_addr),ntohs(servAddr.sin_port));		
	sprintf(msg, "test");
	n=sendto(testfd, msg, strlen(msg), flags, (struct sockaddr *)&servAddr, sizeof(servAddr));
	if(n<0){
		perror("sendto");
		return -1;		
	}
	/* Start test */
	sleep(timeout);
	while(1){
	
		printf("Sending ping %d to %s:%u\n",timeout, inet_ntoa(servAddr.sin_addr),ntohs(servAddr.sin_port));		
		printf("Sending ping %d\n",timeout);		
		sprintf(msg, "ping %d",timeout);
		n=sendto(contfd, msg, strlen(msg), flags, (struct sockaddr *)&servAddr, sizeof(servAddr));
		if(n<0){
			perror("sendto");
			break;
		}
		alarm(1);
		n=recvfrom(testfd,recline,MAX_MSG,flags, (struct sockaddr *)&servAddr,&servLen);
		alarm(0);
		recline[n]=0;
		if(n>0){
			printf("Received %s from %s:%u\n",recline,inet_ntoa(servAddr.sin_addr),
					ntohs(servAddr.sin_port));
			
			close(testfd);
		
			testfd = socket(AF_INET,SOCK_DGRAM,0);
		        if(testfd<0){
				printf("Initial timeout test: cannot reopen test socket\n");
			        perror("socket()");
				return -1;
			}
			
			if(bind(testfd, (struct sockaddr *) &testAddr, sizeof(testAddr))<0){
				perror("bind()");
				return -1;
			}
			

	
			/* Register test address */
			printf("Sending ping %d to %s:%u\n",timeout, inet_ntoa(servAddr.sin_addr),ntohs(servAddr.sin_port));		
			sprintf(msg, "test");
			n=sendto(testfd, msg, strlen(msg), flags, (struct sockaddr *)&servAddr, sizeof(servAddr));
			if(n<0){
				perror("sendto");
				return -1;		
			}
			timeout+=1;
			sleep(timeout);

		}else
			break;
		if(timeout>300){
			printf("No timeout detected within 300 seconds\n");;
			timeout=-1;
			return -1;
		}
	}	
	printf("Initial timeout for %s detected at %d\n",inet_ntoa(servAddr.sin_addr),timeout);
	return timeout-1;

}		

/* initial timeout, with binary search */
int test6(int contfd, int testfd, struct sockaddr_in servAddr, socklen_t servLen, struct sockaddr_in testAddr, int testLen, int start){
	
	int timeout = start;
	int flags = 0;
	int n;
	int inc = start;
	int lo, hi;
	
	lo = start;
	hi = INT_MAX;

	testLen=testLen; /* shutup gcc! */
	 
	printf("Starting timeout test with increment of %d, no ack.\n", inc);
	
	/* Register test address */
	printf("Sending \"test\" %d to %s:%u\n",timeout, inet_ntoa(servAddr.sin_addr),ntohs(servAddr.sin_port));		
	sprintf(msg, "test");
	n=sendto(testfd, msg, strlen(msg), flags, (struct sockaddr *)&servAddr, sizeof(servAddr));
	if(n<0){
		perror("sendto");
		return -1;		
	}
	/* Start test */
	/* give server 1 sec to receive test */
	sleep(1);
	if(timeout>0)
		timeout=timeout-1;
	while(hi - lo  > 1){
	
		printf("sleeping %ds\n", timeout);
		sleep(timeout);
		printf("Sending ping %d to %s:%u\n",timeout, inet_ntoa(servAddr.sin_addr),ntohs(servAddr.sin_port));		
		printf("Sending ping %d\n",timeout);		
		sprintf(msg, "ping %d",timeout);
		n=sendto(contfd, msg, strlen(msg), flags, (struct sockaddr *)&servAddr, sizeof(servAddr));
		if(n<0){
			perror("sendto");
			break;
		}
		alarm(1);
		n=recvfrom(testfd,recline,MAX_MSG,flags, (struct sockaddr *)&servAddr,&servLen);
		alarm(0);
		recline[n]=0;
		if(n>0){
			
			printf("Received %s from %s:%u\n",recline,inet_ntoa(servAddr.sin_addr),
					ntohs(servAddr.sin_port));
			lo = timeout;

			close(testfd);
		
			testfd = socket(AF_INET,SOCK_DGRAM,0);
		        if(testfd<0){
				printf("Initial timeout test: cannot reopen test socket\n");
			        perror("socket()");
				return -1;
			}
			
			if(bind(testfd, (struct sockaddr *) &testAddr, sizeof(testAddr))<0){
				perror("bind()");
				return -1;
			}
			

	
			/* Register test address */
			printf("Sending test %d to %s:%u\n",timeout, inet_ntoa(servAddr.sin_addr),ntohs(servAddr.sin_port));		
			sprintf(msg, "test");
			n=sendto(testfd, msg, strlen(msg), flags, (struct sockaddr *)&servAddr, sizeof(servAddr));
			if(n<0){
				perror("sendto");
				return -1;		
			}
		/*	timeout+=1;*/

		} else {
		        printf("timeout detected for %s at %d seconds\n",inet_ntoa(servAddr.sin_addr), timeout);
			hi = timeout;

			close(testfd);
		
			testfd = socket(AF_INET,SOCK_DGRAM,0);
		        if(testfd<0){
				printf("Initial timeout test: cannot reopen test socket\n");
			        perror("socket()");
				return -1;
			}
			
			if(bind(testfd, (struct sockaddr *) &testAddr, sizeof(testAddr))<0){
				perror("bind()");
				return -1;
			}
			

	
			/* Register test address */
			printf("Sending test %d to %s:%u\n",timeout, inet_ntoa(servAddr.sin_addr),ntohs(servAddr.sin_port));		
			sprintf(msg, "test");
			n=sendto(testfd, msg, strlen(msg), flags, (struct sockaddr *)&servAddr, sizeof(servAddr));
			if(n<0){
				perror("sendto");
				return -1;		
			}


		}
	
		if(hi == INT_MAX){
			if(timeout <= 0){
				timeout = 1;
			} else {
				timeout = timeout * 2;
			}
		} else {
			timeout = (hi + lo + 1) / 2;
		}
	
	
	
	
		if(timeout>=8192){
			printf("No timeout detected within 4096 seconds\n");;
			timeout=-1;
			return -1;
		}
	}	
	printf("Initial timeout for %s detected at %d\n",inet_ntoa(servAddr.sin_addr),timeout);
	return (hi + lo + 1)/2 ;

}	

/* Basic timeout test without ack, binary search */
int test7(int contfd, int testfd, struct sockaddr_in servAddr, socklen_t servLen, int incr){
	
	int timeout = 0;
	int flags = 0;
	int n;
	int inc = incr;
	int lo, hi;
	

	lo = incr;
	hi = INT_MAX;
	printf("Starting timeout test with increment of %d, ack.\n", inc);
	
	/* Register test address */
	printf("Sending \"test\" to %s:%u, timeout=%ds\n", inet_ntoa(servAddr.sin_addr),ntohs(servAddr.sin_port), timeout);		
	
	sprintf(msg, "test");
	n=sendto(testfd, msg, strlen(msg), flags, (struct sockaddr *)&servAddr, sizeof(servAddr));
	if(n<0){
		perror("sendto");
		return -1;		
	}
	/* Start test */
	sleep(1);
	if(timeout>0)
		timeout=timeout-1;
	while(hi - lo > 1)
	{
		printf("sleeping %ds\n", timeout);
		sleep(timeout);
	
		printf("Sending ping %d to %s:%u\n",timeout, inet_ntoa(servAddr.sin_addr),ntohs(servAddr.sin_port));		
		sprintf(msg, "ping %d",timeout);
		n=sendto(contfd, msg, strlen(msg), flags, (struct sockaddr *)&servAddr, sizeof(servAddr));
		if(n<0){
			perror("sendto");
			break;
		}
		alarm(1);
		n=recvfrom(testfd,recline,MAX_MSG,flags, (struct sockaddr *)&servAddr,&servLen);
		alarm(0);
		recline[n]=0;
		if(n>0){
rxok:
			printf("Received %s from %s:%u\n",recline,inet_ntoa(servAddr.sin_addr),
					ntohs(servAddr.sin_port));
			lo = timeout;
			/*
			sprintf(msg,"ack");
			n=sendto(testfd, msg, strlen(msg),flags,(struct sockaddr *)&servAddr, sizeof(servAddr));
			*/
		}else{
			printf("timeout detected for %s at %d seconds\n",inet_ntoa(servAddr.sin_addr), timeout);
			printf("Trying retransmit\n");

			printf("Sending ping %d to %s:%u\n",timeout, inet_ntoa(servAddr.sin_addr),ntohs(servAddr.sin_port));		
			sprintf(msg, "ping %d",timeout);
			n=sendto(contfd, msg, strlen(msg), flags, (struct sockaddr *)&servAddr, sizeof(servAddr));
			if(n<0){
				perror("sendto");
				break;
			}
			alarm(1);
			n=recvfrom(testfd,recline,MAX_MSG,flags, (struct sockaddr *)&servAddr,&servLen);
			alarm(0);
			recline[n]=0;
			if(n>0){
				goto rxok;
			} else {
				hi = timeout;
				/* re-register test address */
				printf("Refreshing \"test\" to %s:%u timeout=%ds\n", inet_ntoa(servAddr.sin_addr),ntohs(servAddr.sin_port), timeout);
				sprintf(msg, "test");
				n=sendto(testfd, msg, strlen(msg), flags, (struct sockaddr *)&servAddr, sizeof(servAddr));
				if(n<0){
					perror("sendto");
					return -1;		
				}
			}
		}
		if(hi == INT_MAX){
			if(timeout <= 0){
				timeout = 1;
			} else {
				timeout = timeout * 2;
			}
		} else {
			timeout = (hi + lo + 1) / 2;
		}
		if(timeout>4096){
			printf("No timeout detected within 4096 seconds\n");;
			timeout=-1;
			return -1;
		}
	}
	printf("Timeout detected between %d and %d (timeout=%ds) for %s\n", lo, hi, timeout,
			inet_ntoa(servAddr.sin_addr));
	return (hi + lo + 1) / 2;
}
		
/* Basic timeout test with ack, binary search */
int test8(int contfd, int testfd, struct sockaddr_in servAddr, socklen_t servLen, int incr){
	
	int timeout = 0;
	int flags = 0;
	int n;
	int inc = incr;
	int lo, hi;
	

	lo = incr;
	hi = INT_MAX;
	printf("Starting timeout test with increment of %d, ack.\n", inc);
	
	/* Register test address */
	printf("Sending \"test\" to %s:%u, timeout=%ds\n", inet_ntoa(servAddr.sin_addr),ntohs(servAddr.sin_port), timeout);		
	
	sprintf(msg, "test");
	n=sendto(testfd, msg, strlen(msg), flags, (struct sockaddr *)&servAddr, sizeof(servAddr));
	if(n<0){
		perror("sendto");
		return -1;		
	}
	/* Start test */
	sleep(1);
	if(timeout>0)
		timeout=timeout-1;
	while(hi - lo > 1)
	{
		printf("sleeping %ds\n", timeout);
		sleep(timeout);
	
		printf("Sending ping %d to %s:%u\n",timeout, inet_ntoa(servAddr.sin_addr),ntohs(servAddr.sin_port));		
		sprintf(msg, "ping %d",timeout);
		n=sendto(contfd, msg, strlen(msg), flags, (struct sockaddr *)&servAddr, sizeof(servAddr));
		if(n<0){
			perror("sendto");
			break;
		}
		alarm(1);
		n=recvfrom(testfd,recline,MAX_MSG,flags, (struct sockaddr *)&servAddr,&servLen);
		alarm(0);
		recline[n]=0;
		if(n>0){
rxok:
			printf("Received %s from %s:%u\n",recline,inet_ntoa(servAddr.sin_addr),
					ntohs(servAddr.sin_port));
			lo = timeout;
			
			sprintf(msg,"ack");
			n=sendto(testfd, msg, strlen(msg),flags,(struct sockaddr *)&servAddr, sizeof(servAddr));
			
		}else{
			printf("timeout detected for %s at %d seconds\n",inet_ntoa(servAddr.sin_addr), timeout);
			printf("Trying retransmit\n");

			printf("Sending ping %d to %s:%u\n",timeout, inet_ntoa(servAddr.sin_addr),ntohs(servAddr.sin_port));		
			sprintf(msg, "ping %d",timeout);
			n=sendto(contfd, msg, strlen(msg), flags, (struct sockaddr *)&servAddr, sizeof(servAddr));
			if(n<0){
				perror("sendto");
				break;
			}
			alarm(1);
			n=recvfrom(testfd,recline,MAX_MSG,flags, (struct sockaddr *)&servAddr,&servLen);
			alarm(0);
			recline[n]=0;
			if(n>0){
				goto rxok;
			} else {
				hi = timeout;
				/* re-register test address */
				printf("Refreshing \"test\" to %s:%u timeout=%ds\n", inet_ntoa(servAddr.sin_addr),ntohs(servAddr.sin_port), timeout);
				sprintf(msg, "test");
				n=sendto(testfd, msg, strlen(msg), flags, (struct sockaddr *)&servAddr, sizeof(servAddr));
				if(n<0){
					perror("sendto");
					return -1;		
				}
			}
		}
		if(hi == INT_MAX){
			if(timeout <= 0){
				timeout = 1;
			} else {
				timeout = timeout * 2;
			}
		} else {
			timeout = (hi + lo + 1) / 2;
		}
		if(timeout>8192){
			printf("No timeout detected within 8192 seconds\n");;
			timeout=-1;
			return -1;
		}
	}
	printf("Timeout detected between %d and %d (timeout=%ds) for %s\n", lo, hi, timeout,
			inet_ntoa(servAddr.sin_addr));
	return (hi + lo + 1) / 2;
}
		


/* Basic timeout test without ack, binary search , traffic */
int test9(int contfd, int testfd, struct sockaddr_in servAddr, socklen_t servLen, int incr){
	
	int timeout = 0;
	int flags = 0;
	int n;
	int inc = incr;
	int lo, hi;
	

	lo = 0;
	hi = INT_MAX;
	printf("Starting timeout test with increment of %d, ack.\n", inc);
	
	/* Register test address */
	printf("Sending \"test\" to %s:%u, timeout=%ds\n", inet_ntoa(servAddr.sin_addr),ntohs(servAddr.sin_port), timeout);		
	
	sprintf(msg, "test");
	n=sendto(testfd, msg, strlen(msg), flags, (struct sockaddr *)&servAddr, sizeof(servAddr));
	if(n<0){
		perror("sendto");
		return -1;		
	}
	
	sleep(1);
	if(timeout>0)
		timeout=timeout-1;
	int i=0;
	printf("Sending traffic to and from testclient for 60 seconds.\n");
	for(i=0;i<60;i++){
		sprintf(msg,"Lorem ipsum dolor sit amet, %d",i);
			
		n=sendto(testfd, msg, strlen(msg), flags, (struct sockaddr *)&servAddr, sizeof(servAddr));
		if(n<0){
			perror("sendto");
			return -1;		
		}
	
		sprintf(msg, "consectetur adipisicing elit, %d",i);
		n=sendto(contfd, msg, strlen(msg), flags, (struct sockaddr *)&servAddr, sizeof(servAddr));
		if(n<0){
			perror("sendto");
			break;
		}
	//	n=recvfrom(testfd,recline,MAX_MSG,flags, (struct sockaddr *)&servAddr,&servLen);
		sleep(1);
	}
	
	
	/* Start test */
	while(hi - lo > 1)
	{
		printf("sleeping %ds\n", timeout);
		sleep(timeout);
	
		printf("Sending ping %d to %s:%u\n",timeout, inet_ntoa(servAddr.sin_addr),ntohs(servAddr.sin_port));		
		sprintf(msg, "ping %d",timeout);
		n=sendto(contfd, msg, strlen(msg), flags, (struct sockaddr *)&servAddr, sizeof(servAddr));
		if(n<0){
			perror("sendto");
			break;
		}
		alarm(1);
		n=recvfrom(testfd,recline,MAX_MSG,flags, (struct sockaddr *)&servAddr,&servLen);
		alarm(0);
		recline[n]=0;
		if(n>0){
rxok:
			printf("Received %s from %s:%u\n",recline,inet_ntoa(servAddr.sin_addr),
					ntohs(servAddr.sin_port));
			lo = timeout;
	/*		
			sprintf(msg,"ack");
			n=sendto(testfd, msg, strlen(msg),flags,(struct sockaddr *)&servAddr, sizeof(servAddr));
	*/		
		}else{
			printf("timeout detected for %s at %d seconds\n",inet_ntoa(servAddr.sin_addr), timeout);
			printf("Trying retransmit\n");

			printf("Sending ping %d to %s:%u\n",timeout, inet_ntoa(servAddr.sin_addr),ntohs(servAddr.sin_port));		
			sprintf(msg, "ping %d",timeout);
			n=sendto(contfd, msg, strlen(msg), flags, (struct sockaddr *)&servAddr, sizeof(servAddr));
			if(n<0){
				perror("sendto");
				break;
			}
			alarm(1);
			n=recvfrom(testfd,recline,MAX_MSG,flags, (struct sockaddr *)&servAddr,&servLen);
			alarm(0);
			recline[n]=0;
			if(n>0){
				goto rxok;
			} else {
				hi = timeout;
				/* re-register test address */
				printf("Refreshing \"test\" to %s:%u timeout=%ds\n", inet_ntoa(servAddr.sin_addr),ntohs(servAddr.sin_port), timeout);
				sprintf(msg, "test");
				n=sendto(testfd, msg, strlen(msg), flags, (struct sockaddr *)&servAddr, sizeof(servAddr));
				if(n<0){
					perror("sendto");
					return -1;		
				}
			}
		}
		if(hi == INT_MAX){
			if(timeout <= 0){
				timeout = 1;
			} else {
				timeout = timeout * 2;
			}
		} else {
			timeout = (hi + lo + 1) / 2;
		}
		if(timeout>4096){
			printf("No timeout detected within 4096 seconds\n");;
			timeout=-1;
			return -1;
		}
	}
	printf("Timeout detected between %d and %d (timeout=%ds) for %s\n", lo, hi, timeout,
			inet_ntoa(servAddr.sin_addr));
	return (hi + lo + 1) / 2;
}


/* Basic timeout test with ack, binary search , traffic */
int test10(int contfd, int testfd, struct sockaddr_in servAddr, socklen_t servLen, int incr){
	
	int timeout = 0;
	int flags = 0;
	int n;
	int inc = incr;
	int lo, hi;
	

	lo = 0;
	hi = INT_MAX;
	printf("Starting timeout test with increment of %d, ack.\n", inc);
	
	/* Register test address */
	printf("Sending \"test\" to %s:%u, timeout=%ds\n", inet_ntoa(servAddr.sin_addr),ntohs(servAddr.sin_port), timeout);		
	
	sprintf(msg, "test");
	n=sendto(testfd, msg, strlen(msg), flags, (struct sockaddr *)&servAddr, sizeof(servAddr));
	if(n<0){
		perror("sendto");
		return -1;		
	}
	
	int i=0;
	printf("Sending traffic to and from testclient for 60 seconds.\n");
	for(i=0;i<60;i++){
		sprintf(msg,"Lorem ipsum dolor sit amet, %d",i);
			
		n=sendto(testfd, msg, strlen(msg), flags, (struct sockaddr *)&servAddr, sizeof(servAddr));
		if(n<0){
			perror("sendto");
			return -1;		
		}
	
		sprintf(msg, "consectetur adipisicing elit, %d",i);
		n=sendto(contfd, msg, strlen(msg), flags, (struct sockaddr *)&servAddr, sizeof(servAddr));
		if(n<0){
			perror("sendto");
			break;
		}
	//	n=recvfrom(testfd,recline,MAX_MSG,flags, (struct sockaddr *)&servAddr,&servLen);
		sleep(1);
	}
	
	
	/* Start test */
	sleep(1);
	if(timeout>0)
		timeout=timeout-1;
	while(hi - lo > 1)
	{
		printf("sleeping %ds\n", timeout);
		sleep(timeout);
	
		printf("Sending ping %d to %s:%u\n",timeout, inet_ntoa(servAddr.sin_addr),ntohs(servAddr.sin_port));		
		sprintf(msg, "ping %d",timeout);
		n=sendto(contfd, msg, strlen(msg), flags, (struct sockaddr *)&servAddr, sizeof(servAddr));
		if(n<0){
			perror("sendto");
			break;
		}
		alarm(1);
		n=recvfrom(testfd,recline,MAX_MSG,flags, (struct sockaddr *)&servAddr,&servLen);
		alarm(0);
		recline[n]=0;
		if(n>0){
rxok:
			printf("Received %s from %s:%u\n",recline,inet_ntoa(servAddr.sin_addr),
					ntohs(servAddr.sin_port));
			lo = timeout;
			
			sprintf(msg,"ack");
			n=sendto(testfd, msg, strlen(msg),flags,(struct sockaddr *)&servAddr, sizeof(servAddr));
			
		}else{
			printf("timeout detected for %s at %d seconds\n",inet_ntoa(servAddr.sin_addr), timeout);
			printf("Trying retransmit\n");

			printf("Sending ping %d to %s:%u\n",timeout, inet_ntoa(servAddr.sin_addr),ntohs(servAddr.sin_port));		
			sprintf(msg, "ping %d",timeout);
			n=sendto(contfd, msg, strlen(msg), flags, (struct sockaddr *)&servAddr, sizeof(servAddr));
			if(n<0){
				perror("sendto");
				break;
			}
			alarm(1);
			n=recvfrom(testfd,recline,MAX_MSG,flags, (struct sockaddr *)&servAddr,&servLen);
			alarm(0);
			recline[n]=0;
			if(n>0){
				goto rxok;
			} else {
				hi = timeout;
				/* re-register test address */
				printf("Refreshing \"test\" to %s:%u timeout=%ds\n", inet_ntoa(servAddr.sin_addr),ntohs(servAddr.sin_port), timeout);
				sprintf(msg, "test");
				n=sendto(testfd, msg, strlen(msg), flags, (struct sockaddr *)&servAddr, sizeof(servAddr));
				if(n<0){
					perror("sendto");
					return -1;		
				}
			}
		}
		if(hi == INT_MAX){
			if(timeout <= 0){
				timeout = 1;
			} else {
				timeout = timeout * 2;
			}
		} else {
			timeout = (hi + lo + 1) / 2;
		}
		if(timeout>4096){
			printf("No timeout detected within 4096 seconds\n");;
			timeout=-1;
			return -1;
		}
	}
	printf("Timeout detected between %d and %d (timeout=%ds) for %s\n", lo, hi, timeout,
			inet_ntoa(servAddr.sin_addr));
	return (hi + lo + 1) / 2;
}



/* Blocking timer test */
int test11(int contfd, int testfd, struct sockaddr_in servAddr, socklen_t servLen, int incr){
	
	int timeout = 0;
	int flags = 0;
	int n;
	int inc = incr;
	int lo, hi;
  int retrans = 0;	

	lo = incr;
	hi = INT_MAX;
	printf("Starting timeout test with increment of %d, ack.\n", inc);
	
	/* Register test address */
	printf("Sending \"test\" to %s:%u, timeout=%ds\n", inet_ntoa(servAddr.sin_addr),ntohs(servAddr.sin_port), timeout);		
	
	sprintf(msg, "test");
	n=sendto(testfd, msg, strlen(msg), flags, (struct sockaddr *)&servAddr, sizeof(servAddr));
	if(n<0){
		perror("sendto");
		return -1;		
	}
	/* Start test */
	sleep(1);
	if(timeout>0)
		timeout=timeout-1;
	while(hi - lo > 1)
	{
		printf("sleeping %ds\n", timeout);
		sleep(timeout);
	
		printf("Sending ping %d to %s:%u\n",timeout, inet_ntoa(servAddr.sin_addr),ntohs(servAddr.sin_port));		
		sprintf(msg, "ping %d",timeout);
		n=sendto(contfd, msg, strlen(msg), flags, (struct sockaddr *)&servAddr, sizeof(servAddr));
		if(n<0){
			perror("sendto");
			break;
		}
		alarm(1);
		n=recvfrom(testfd,recline,MAX_MSG,flags, (struct sockaddr *)&servAddr,&servLen);
		alarm(0);
		recline[n]=0;
		if(n>0){
rxok:
			printf("Received %s from %s:%u\n",recline,inet_ntoa(servAddr.sin_addr),
					ntohs(servAddr.sin_port));
			lo = timeout;
			/*
			sprintf(msg,"ack");
			n=sendto(testfd, msg, strlen(msg),flags,(struct sockaddr *)&servAddr, sizeof(servAddr));
			*/
		}else{
rxfail:
			printf("timeout detected for %s at %d seconds\n",inet_ntoa(servAddr.sin_addr), timeout);
			printf("Trying retransmit\n");
			retrans++;
			printf("Sending ping %d to %s:%u\n",timeout, inet_ntoa(servAddr.sin_addr),ntohs(servAddr.sin_port));		
			sprintf(msg, "ping %d",timeout);
			n=sendto(contfd, msg, strlen(msg), flags, (struct sockaddr *)&servAddr, sizeof(servAddr));
			if(n<0){
				perror("sendto");
				break;
			}
			alarm(1);
			n=recvfrom(testfd,recline,MAX_MSG,flags, (struct sockaddr *)&servAddr,&servLen);
			alarm(0);
			recline[n]=0;
			if(n>0){
				goto rxok;
			} else {
				if(retrans==1)
					goto rxfail;
				hi = timeout;
				/* re-register test address */
				printf("Refreshing \"test\" to %s:%u timeout=%ds\n", inet_ntoa(servAddr.sin_addr),ntohs(servAddr.sin_port), timeout);
				sprintf(msg, "test");
				n=sendto(testfd, msg, strlen(msg), flags, (struct sockaddr *)&servAddr, sizeof(servAddr));
				if(n<0){
					perror("sendto");
					return -1;		
				}
			}
		}
		if(hi == INT_MAX){
			if(timeout <= 0){
				timeout = 1;
			} else {
				timeout = timeout * 2;
			}
		} else {
			timeout = (hi + lo + 1) / 2;
		}
		if(timeout>4096){
			printf("No timeout detected within 4096 seconds\n");;
			timeout=-1;
			return -1;
		}

	}
	
	
	
	
	
	printf("Timeout detected between %d and %d (timeout=%ds) for %s\n", lo, hi, timeout,
			inet_ntoa(servAddr.sin_addr));
	
	printf("Starting check for blocking timer\n");
	

	 printf("Refreshing control channel\n");
	 sprintf(msg, "ref");
	 n=sendto(contfd, msg, strlen(msg), flags, (struct sockaddr *)&servAddr, sizeof(servAddr));
	 if(n<0){
		 perror("sendto");
		 return -1;
	 }

	sprintf(msg,"test");
	n=sendto(testfd, msg, strlen(msg), flags, (struct sockaddr *)&servAddr, sizeof(servAddr));
	if(n<0){
		perror("sendto");
		return -1;		
	}
	timeout=timeout+1;
	sleep(timeout);

	printf("Refreshing control channel\n");
	sprintf(msg, "ref");
	n=sendto(contfd, msg, strlen(msg), flags, (struct sockaddr *)&servAddr, sizeof(servAddr));
	if(n<0){
		perror("sendto");
		return -1;
	}

	alarm(1);
	n=recvfrom(contfd,recline,MAX_MSG,flags, (struct sockaddr *)&servAddr,&servLen);
	n=recvfrom(contfd,recline,MAX_MSG,flags, (struct sockaddr *)&servAddr,&servLen);
	n=recvfrom(contfd,recline,MAX_MSG,flags, (struct sockaddr *)&servAddr,&servLen);
	alarm(0);
	
	int i;
	for(i=0;i<10;i++){
		printf("Sending block %d to %s:%u\n",timeout+i, inet_ntoa(servAddr.sin_addr),ntohs(servAddr.sin_port));		
		sprintf(msg, "block %d",timeout+i);
		n=sendto(testfd, msg, strlen(msg), flags, (struct sockaddr *)&servAddr, sizeof(servAddr));
		if(n<0){
			perror("sendto");
			break;
		}
		alarm(1);
		n=recvfrom(contfd,recline,MAX_MSG,flags, (struct sockaddr *)&servAddr,&servLen);
		alarm(0);
		recline[n]=0;
		if(n<0){
			printf("No message received at %d\n",timeout+i);
			continue;
		} else {
			printf("recline is: %s\n",recline);
			if(strncmp(recline,"block",5)==0){
				if (i==0){
					printf("No blocking timer");
					return -1;
				}
				printf("blocking timer at %d\n",timeout+i);
				return timeout;
			}else{
				printf("Binding changed\n");
				return -1;
			}
		}
		


	}
	printf("No blocking timer detected\n");
	return 0;
}

		

/* initial timeout, with incremental search, modified for defined ports  */
int test21(int contfd, int testfd, struct sockaddr_in servAddr, socklen_t servLen, struct sockaddr_in testAddr, int testLen, int start){
	
	int timeout = start-5;
	int flags = 0;
	int n;
	int inc = 1;

	testLen=testLen; /* shutup gcc! */
	printf("Starting initial timeout test with increment of %d, from %d.\n", inc,timeout);
	
	/* Register test address */
	printf("Sending \"test\" %d to %s:%u\n",timeout, inet_ntoa(servAddr.sin_addr),ntohs(servAddr.sin_port));		
	sprintf(msg, "test");
	n=sendto(testfd, msg, strlen(msg), flags, (struct sockaddr *)&servAddr, sizeof(servAddr));
	if(n<0){
		perror("sendto");
		return -1;		
	}
	/* Start test */
	sleep(timeout);
	while(timeout<(start+5)){
	
		printf("Sending ping %d to %s:%u\n",timeout, inet_ntoa(servAddr.sin_addr),ntohs(servAddr.sin_port));		
		printf("Sending ping %d\n",timeout);		
		sprintf(msg, "ping %d",timeout);
		n=sendto(contfd, msg, strlen(msg), flags, (struct sockaddr *)&servAddr, sizeof(servAddr));
		if(n<0){
			perror("sendto");
			break;
		}
		alarm(1);
		n=recvfrom(testfd,recline,MAX_MSG,flags, (struct sockaddr *)&servAddr,&servLen);
		alarm(0);
		recline[n]=0;
		if(n>0){
			printf("Received %s from %s:%u\n",recline,inet_ntoa(servAddr.sin_addr),
					ntohs(servAddr.sin_port));
			
			close(testfd);
		
			testfd = socket(AF_INET,SOCK_DGRAM,0);
		        if(testfd<0){
				printf("Initial timeout test: cannot reopen test socket\n");
			        perror("socket()");
				return -1;
			}
			
			if(bind(testfd, (struct sockaddr *) &testAddr, sizeof(testAddr))<0){
				perror("bind()");
				return -1;
			}
			

	
			/* Register test address */
			printf("Sending ping %d to %s:%u\n",timeout, inet_ntoa(servAddr.sin_addr),ntohs(servAddr.sin_port));		
			sprintf(msg, "test");
			n=sendto(testfd, msg, strlen(msg), flags, (struct sockaddr *)&servAddr, sizeof(servAddr));
			if(n<0){
				perror("sendto");
				return -1;		
			}
			timeout+=1;
			sleep(timeout);

		}else
			break;
	}	
	printf("Timeout detected for %s:%u detected at %d\n",inet_ntoa(servAddr.sin_addr),ntohs(servAddr.sin_port),timeout);
	return timeout-1;

}		


/* Basic timeout test without ack, binary search. Use this, earlier binary tests had bugs */
int test22(int contfd, int testfd, struct sockaddr_in servAddr, socklen_t servLen, int incr){
	
	int timeout = 1;
	int flags = 0;
	int n;
	int inc = incr;
	int lo, hi;
	

	lo = incr;
	hi = 86400;
	printf("Starting timeout test with increment of %d, ack.\n", inc);
	
	/* Register test address */
	printf("Sending \"test\" to %s:%u, timeout=%ds\n", inet_ntoa(servAddr.sin_addr),ntohs(servAddr.sin_port), timeout);		
	
	sprintf(msg, "test");
	n=sendto(testfd, msg, strlen(msg), flags, (struct sockaddr *)&servAddr, sizeof(servAddr));
	if(n<0){
		perror("sendto");
		return -1;		
	}
	/* Start test */
	while(hi - lo > 1)
	{
		printf("sleeping %ds\n", timeout);
		sleep(timeout);
	
		printf("Sending ping %d to %s:%u\n",timeout, inet_ntoa(servAddr.sin_addr),ntohs(servAddr.sin_port));		
		sprintf(msg, "ping %d",timeout);
		n=sendto(contfd, msg, strlen(msg), flags, (struct sockaddr *)&servAddr, sizeof(servAddr));
		if(n<0){
			perror("sendto");
			break;
		}
		alarm(1);
		n=recvfrom(testfd,recline,MAX_MSG,flags, (struct sockaddr *)&servAddr,&servLen);
		alarm(0);
		recline[n]=0;
		if(n>0){
rxok:
			printf("Received %s from %s:%u\n",recline,inet_ntoa(servAddr.sin_addr),
					ntohs(servAddr.sin_port));
			if(timeout>lo){
				lo = timeout;
			} 
			
			timeout = timeout * 2; 
			
					
			if(timeout >= hi)
				timeout = (hi + lo + 1)/2;


			/*
			sprintf(msg,"ack");
			n=sendto(testfd, msg, strlen(msg),flags,(struct sockaddr *)&servAddr, sizeof(servAddr));
			*/
		}else{
			printf("timeout detected for %s at %d seconds\n",inet_ntoa(servAddr.sin_addr), timeout);
			printf("Trying retransmit\n");

			printf("Sending ping %d to %s:%u\n",timeout, inet_ntoa(servAddr.sin_addr),ntohs(servAddr.sin_port));		
			sprintf(msg, "ping %d",timeout);
			n=sendto(contfd, msg, strlen(msg), flags, (struct sockaddr *)&servAddr, sizeof(servAddr));
			if(n<0){
				perror("sendto");
				break;
			}
			alarm(1);
			n=recvfrom(testfd,recline,MAX_MSG,flags, (struct sockaddr *)&servAddr,&servLen);
			alarm(0);
			recline[n]=0;
			if(n>0){
				goto rxok;
			} else {
				hi = timeout;
				timeout=1;

				/* re-register test address */
				printf("Refreshing \"test\" to %s:%u timeout=%ds\n", inet_ntoa(servAddr.sin_addr),ntohs(servAddr.sin_port), timeout);
				sprintf(msg, "test");
				n=sendto(testfd, msg, strlen(msg), flags, (struct sockaddr *)&servAddr, sizeof(servAddr));
				if(n<0){
					perror("sendto");
					return -1;		
				}
			}
		}


		/*if(hi == INT_MAX){
			if(timeout <= 0){
				timeout = 1;
			} else {
				timeout = timeout * 2;
			}
		} else {
			timeout = (hi + lo + 1) / 2;

		}
		*/
		//timeout = timeout * 2;




/*
		if(timeout>4096){
			printf("No timeout detected within 4096 seconds\n");;
			timeout=-1;
			return -1;
		}*/

	}
	printf("Timeout detected between %d and %d (timeout=%ds) for %s\n", lo, hi, timeout,
			inet_ntoa(servAddr.sin_addr));
	return (hi + lo + 1) / 2;
}



/* Basic timeout test with ack, binary search, use this, earlier binary tests had bugs */
int test23(int contfd, int testfd, struct sockaddr_in servAddr, socklen_t servLen, int incr){
	
	int timeout = 1;
	int flags = 0;
	int n;
	int inc = incr;
	int lo, hi;
	

	lo = incr;
	hi = 86400;
	printf("Starting timeout test with increment of %d, ack.\n", inc);
	
	/* Register test address */
	printf("Sending \"test\" to %s:%u, timeout=%ds\n", inet_ntoa(servAddr.sin_addr),ntohs(servAddr.sin_port), timeout);		
	
	sprintf(msg, "test");
	n=sendto(testfd, msg, strlen(msg), flags, (struct sockaddr *)&servAddr, sizeof(servAddr));
	if(n<0){
		perror("sendto");
		return -1;		
	}
	/* Start test */
	while(hi - lo > 1)
	{
		printf("sleeping %ds\n", timeout);
		sleep(timeout);
	
		printf("Sending ping %d to %s:%u\n",timeout, inet_ntoa(servAddr.sin_addr),ntohs(servAddr.sin_port));		
		sprintf(msg, "ping %d",timeout);
		n=sendto(contfd, msg, strlen(msg), flags, (struct sockaddr *)&servAddr, sizeof(servAddr));
		if(n<0){
			perror("sendto");
			break;
		}
		alarm(1);
		n=recvfrom(testfd,recline,MAX_MSG,flags, (struct sockaddr *)&servAddr,&servLen);
		alarm(0);
		recline[n]=0;
		if(n>0){
rxok:
			printf("Received %s from %s:%u\n",recline,inet_ntoa(servAddr.sin_addr),
					ntohs(servAddr.sin_port));
			if(timeout>lo){
				lo = timeout;
			} 
			
			timeout = timeout * 2; 
			
					
			if(timeout >= hi)
				timeout = (hi + lo + 1)/2;


			printf("Sending ack back\n");			
			sprintf(msg,"ack");
			n=sendto(testfd, msg, strlen(msg),flags,(struct sockaddr *)&servAddr, sizeof(servAddr));
			
		}else{
			printf("timeout detected for %s at %d seconds\n",inet_ntoa(servAddr.sin_addr), timeout);
			printf("Trying retransmit\n");

			printf("Sending ping %d to %s:%u\n",timeout, inet_ntoa(servAddr.sin_addr),ntohs(servAddr.sin_port));		
			sprintf(msg, "ping %d",timeout);
			n=sendto(contfd, msg, strlen(msg), flags, (struct sockaddr *)&servAddr, sizeof(servAddr));
			if(n<0){
				perror("sendto");
				break;
			}
			alarm(1);
			n=recvfrom(testfd,recline,MAX_MSG,flags, (struct sockaddr *)&servAddr,&servLen);
			alarm(0);
			recline[n]=0;
			if(n>0){
				goto rxok;
			} else {
				hi = timeout;
				timeout=1;

				/* re-register test address */
				printf("Refreshing \"test\" to %s:%u timeout=%ds\n", inet_ntoa(servAddr.sin_addr),ntohs(servAddr.sin_port), timeout);
				sprintf(msg, "test");
				n=sendto(testfd, msg, strlen(msg), flags, (struct sockaddr *)&servAddr, sizeof(servAddr));
				if(n<0){
					perror("sendto");
					return -1;		
				}
			}
		}


		/*if(hi == INT_MAX){
			if(timeout <= 0){
				timeout = 1;
			} else {
				timeout = timeout * 2;
			}
		} else {
			timeout = (hi + lo + 1) / 2;

		}
		*/
		//timeout = timeout * 2;




/*
		if(timeout>4096){
			printf("No timeout detected within 4096 seconds\n");;
			timeout=-1;
			return -1;
		}*/

	}
	printf("Timeout detected between %d and %d (timeout=%ds) for %s\n", lo, hi, timeout,
			inet_ntoa(servAddr.sin_addr));
	return (hi + lo + 1) / 2;
}


/* Basic timeout test for outgoing traffic, binary search. */
int test24(int contfd, int testfd, struct sockaddr_in servAddr, socklen_t servLen, int incr){
	
	int timeout = 1;
	int flags = 0;
	int n;
	int inc = incr;
	int lo, hi;
	

	lo = incr;
	hi = 86400;
	printf("Starting timeout test with increment of %d, ack.\n", inc);
	
	/* Register test address */
	printf("Sending \"test\" to %s:%u, timeout=%ds\n", inet_ntoa(servAddr.sin_addr),ntohs(servAddr.sin_port), timeout);		
	
	sprintf(msg, "test");
	n=sendto(testfd, msg, strlen(msg), flags, (struct sockaddr *)&servAddr, sizeof(servAddr));
	if(n<0){
		perror("sendto");
		return -1;		
	}
	/* Start test */
	while(hi - lo > 1)
	{
		printf("sleeping %ds\n", timeout);
		

		/* Refresh control channel */
		int refresh;
		refresh = 14;
		while(refresh < timeout){
			sleep(14);
			printf("Refreshing\n");

			refresh = refresh + 15;
			sprintf(msg, "ref");
			n=sendto(contfd, msg, strlen(msg), flags, (struct sockaddr *)&servAddr, sizeof(servAddr));
			alarm(1);
			n=recvfrom(contfd,recline,MAX_MSG,flags, (struct sockaddr *)&servAddr,&servLen);
			alarm(0);
			
			if(n<0){
				perror("sendto ref");
				break;
			}
		}
	
		printf("Sending pong %d to %s:%u\n",timeout, inet_ntoa(servAddr.sin_addr),ntohs(servAddr.sin_port));		
		sprintf(msg, "pong %d",timeout);
		n=sendto(testfd, msg, strlen(msg), flags, (struct sockaddr *)&servAddr, sizeof(servAddr));
		if(n<0){
			perror("sendto");
			break;
		}
		alarm(1);
		n=recvfrom(contfd,recline,MAX_MSG,flags, (struct sockaddr *)&servAddr,&servLen);
		alarm(0);
		recline[n]=0;
		if(n>0){
rxok:
			printf("Received %s from %s:%u\n",recline,inet_ntoa(servAddr.sin_addr),
					ntohs(servAddr.sin_port));
			if(timeout>lo){
				lo = timeout;
			} 
			
			timeout = timeout * 2; 
			
					
			if(timeout >= hi)
				timeout = (hi + lo + 1)/2;


			/*
			sprintf(msg,"ack");
			n=sendto(testfd, msg, strlen(msg),flags,(struct sockaddr *)&servAddr, sizeof(servAddr));
			*/
		}else{
			printf("timeout detected for %s at %d seconds\n",inet_ntoa(servAddr.sin_addr), timeout);
			printf("Trying retransmit\n");

			printf("Sending pong %d to %s:%u\n",timeout, inet_ntoa(servAddr.sin_addr),ntohs(servAddr.sin_port));		
			sprintf(msg, "pong %d",timeout);
			n=sendto(testfd, msg, strlen(msg), flags, (struct sockaddr *)&servAddr, sizeof(servAddr));
			if(n<0){
				perror("sendto");
				break;
			}
			alarm(1);
			n=recvfrom(contfd,recline,MAX_MSG,flags, (struct sockaddr *)&servAddr,&servLen);
			alarm(0);
			recline[n]=0;
			if(n>0){
				goto rxok;
			} else {
				hi = timeout;
				timeout=1;

				/* re-register test address */
				printf("Refreshing \"test\" to %s:%u timeout=%ds\n", inet_ntoa(servAddr.sin_addr),ntohs(servAddr.sin_port), timeout);
				sprintf(msg, "test");
				n=sendto(testfd, msg, strlen(msg), flags, (struct sockaddr *)&servAddr, sizeof(servAddr));
				if(n<0){
					perror("sendto");
					return -1;		
				}
			}
		}


		/*if(hi == INT_MAX){
			if(timeout <= 0){
				timeout = 1;
			} else {
				timeout = timeout * 2;
			}
		} else {
			timeout = (hi + lo + 1) / 2;

		}
		*/
		//timeout = timeout * 2;




/*
		if(timeout>4096){
			printf("No timeout detected within 4096 seconds\n");;
			timeout=-1;
			return -1;
		}*/

	}
	printf("Timeout detected between %d and %d (timeout=%ds) for %s\n", lo, hi, timeout,
			inet_ntoa(servAddr.sin_addr));
	return (hi + lo + 1) / 2;
}
int main(int argc, char *argv[]) {

	int contfd, testfd, n, testLen, flags, timeOut;
	socklen_t servLen;
	struct sockaddr_in contAddr, testAddr, servAddr;
	struct hostent *h;
	struct sigaction act;


	int testnro=0;

	/* check command line args */
	if(argc<4) {
		printf("usage : %s <server> <port> <test> <increment> \n", argv[0]);
		exit(1);
	}
	
	testnro=atoi(argv[3]);

	/* get server IP address (no check if input is IP address or DNS name */
	h = gethostbyname(argv[1]);
	if(h==NULL) {
		printf("%s: unknown host '%s' \n", argv[0], argv[1]);
		exit(1);
	}

	printf("%s: Starting timeout test for  '%s' (IP : %s) \n", argv[0], h->h_name,
	inet_ntoa(*(struct in_addr *)h->h_addr_list[0]));

	servAddr.sin_family = AF_INET;
	memcpy((void *)&servAddr.sin_addr, *h->h_addr_list, h->h_length);
	servAddr.sin_port = htons(atoi(argv[2]));

	/* socket creation */
	contfd = socket(AF_INET,SOCK_DGRAM,0);
	if(contfd<0) {
		printf("%s: cannot open control socket \n",argv[0]);
		exit(1);
	}

	testfd = socket(AF_INET,SOCK_DGRAM,0);
	if(testfd<0){
		printf("%s: cannot open test socket\n",argv[0]);
		exit(1);
	}

	/* bind any port */
	
	contAddr.sin_family = AF_INET;
	contAddr.sin_addr.s_addr = htonl(INADDR_ANY);
	contAddr.sin_port = htons(0);
	testAddr.sin_family = AF_INET;
	testAddr.sin_addr.s_addr = htonl(INADDR_ANY);
	testAddr.sin_port = htons(0);

	if(bind(contfd, (struct sockaddr *) &contAddr, sizeof(contAddr))<0){
		perror("bind");
		exit(1);
	}

	if(bind(testfd, (struct sockaddr *) &testAddr, sizeof(testAddr))<0){
		perror("bind");
		exit(1);
	}

	flags = 0;

	timeOut = 100; // ms

	servLen=sizeof(servAddr);
	testLen=sizeof(testAddr);

       /* set sighand so read() gets interrupted by SIGALRM */
	memset(&act, 0, sizeof act);    
	act.sa_handler = sigalrm;
	sigfillset(&act.sa_mask);
	sigaction(SIGALRM, &act, NULL);

	/* Handshake with server */
	
	memset(msg,0,MAX_MSG);	
	sprintf(msg,"ref");
	
	n=sendto(contfd, msg, strlen(msg), flags, (struct sockaddr *)&servAddr, sizeof(servAddr));

	alarm(1);
	n=recvfrom(contfd,msg,MAX_MSG,flags, (struct sockaddr *)&servAddr,&servLen);
	msg[n]=0;

	printf("%s:%u %s\n",inet_ntoa(servAddr.sin_addr),ntohs(servAddr.sin_port), msg);
	alarm(0);
	
	if(testnro==1){
		
		printf("Starting timeout test-1\n");
		n=test1(contfd,testfd, servAddr, servLen, atoi(argv[4]));
	        printf("Starting timeout test-3 with fine grained search from %d\n",n-2);
		if(n>0)
			n=test3(contfd,testfd,servAddr,servLen,2,n);
		
		printf("Test finished for %s with timeout detected at %d\n",inet_ntoa(servAddr.sin_addr),n);
	}

	if(testnro==2){
		printf("Starting timeout test-2\n");
		n=test2(contfd,testfd, servAddr, servLen, atoi(argv[4]));
		printf("Starting timeout test-3 with fine grained search from %d\n",n); 
		if(n>0)
			n=test3(contfd,testfd,servAddr,servLen,1,n-5);
		printf("Test finished for %s with timeout detected at %d\n",inet_ntoa(servAddr.sin_addr),n);

	}
	
	if(testnro==3){
		n=test3(contfd,testfd, servAddr, servLen, atoi(argv[4]),atoi(argv[5]));
	        if(n>0)                
			n=test4(contfd,testfd,servAddr,servLen,1,n);

	
	}

	if(testnro==4){
		n=test4(contfd,testfd, servAddr, servLen, atoi(argv[4]),atoi(argv[5]));
	        if(n>0)                
			n=test4(contfd,testfd,servAddr,servLen,1,n);

	
	}
	if(testnro==5){
		printf("Starting initial timeout test\n");
		n=test5(contfd,testfd,servAddr,servLen,testAddr,testLen,atoi(argv[4]));
		
	}		
	if(testnro==6){
		printf("Starting initial timeout test, binary search\n");
		n=test6(contfd,testfd,servAddr,servLen,testAddr,testLen,atoi(argv[4]));
		
	}		
	if(testnro==7){
		n=test7(contfd,testfd, servAddr, servLen, atoi(argv[4]));
//	        printf("Starting timeout test-3 with fine grained search from %d\n",n-5);
//		if(n>0)
//			n=test3(contfd,testfd,servAddr,servLen,1,n-5);
		printf("Test finished for %s with timeout detected at %d\n",inet_ntoa(servAddr.sin_addr),n);
		
	}		
	if(testnro==8){
		n=test8(contfd,testfd, servAddr, servLen, atoi(argv[4]));
//	        printf("Starting timeout test-3 with fine grained search from %d\n",n-5);
//		if(n>0)
//			n=test3(contfd,testfd,servAddr,servLen,1,n-5);
		printf("Test finished for %s with timeout detected at %d\n",inet_ntoa(servAddr.sin_addr),n);
		
	}		
	if(testnro==9){
		n=test9(contfd,testfd, servAddr, servLen, atoi(argv[4]));
		printf("Test finished for %s with timeout detected at %d\n",inet_ntoa(servAddr.sin_addr),n);
		
	}		
	if(testnro==10){
		n=test10(contfd,testfd, servAddr, servLen, atoi(argv[4]));
		printf("Test finished for %s with timeout detected at %d\n",inet_ntoa(servAddr.sin_addr),n);
		
	}		

	if(testnro==11){
		n=test11(contfd,testfd, servAddr, servLen, atoi(argv[4]));
		printf("Test finished for %s with blocking timer at %d\n",inet_ntoa(servAddr.sin_addr),n);
		
	}		

	if(testnro==20){
		
		printf("Starting timeout test-1\n");
		n=test1(contfd,testfd, servAddr, servLen, atoi(argv[4]));
		
		if(n<=0){
			printf("Increment probably too big, starting with lower value 20\n");
			n=test1(contfd,testfd, servAddr, servLen, 20);
		}
		if(n>0){
			printf("Starting timeout test-3 with fine grained search from %d\n",n-2);
			n=test3(contfd,testfd,servAddr,servLen,2,n);
		}
		
		
		
		printf("Test finished for %s with timeout detected at %d\n",inet_ntoa(servAddr.sin_addr),n);
	}
	/* Adjusted timeout test (initial value given -5) */ 
	if(testnro==21){
		
		printf("Starting timeout test-21\n");
		n=test21(contfd,testfd,servAddr,servLen,testAddr,testLen,atoi(argv[4]));
		/*
		if(n>0){
			printf("Starting timeout test-3 with fine grained search from %d\n",n-2);
			n=test3(contfd,testfd,servAddr,servLen,2,n);
		}*/
			
		
		
		printf("Test finished for %s with timeout detected at %d\n",inet_ntoa(servAddr.sin_addr),n);
	}


	
	if(testnro==22){
		n=test22(contfd,testfd, servAddr, servLen, atoi(argv[4]));
		printf("Test finished for %s with timeout detected at %d\n",inet_ntoa(servAddr.sin_addr),n);
		
	}		
	if(testnro==23){
		n=test23(contfd,testfd, servAddr, servLen, atoi(argv[4]));
		printf("Test finished for %s with timeout detected at %d\n",inet_ntoa(servAddr.sin_addr),n);
		
	}		
	
	if(testnro==24){
		n=test24(contfd,testfd, servAddr, servLen, atoi(argv[4]));
		printf("Test finished for %s with timeout detected at %d\n",inet_ntoa(servAddr.sin_addr),n);
		
	}		
	FILE *data = fopen("data","w");
	
	fprintf(data,"%s %s %d %d ",argv[1],argv[2],atoi(argv[3]), n);
	fprintf(data,"\n");
	fclose(data);
	
	
	sleep(1);
	printf("Ordering server to quit\n");

	sprintf(msg,"quit");
	
	
	n=sendto(contfd, msg, strlen(msg), flags, (struct sockaddr *)&servAddr, sizeof(servAddr));

	alarm(1);
	n=recvfrom(contfd,msg,MAX_MSG,flags, (struct sockaddr *)&servAddr,&servLen);
	msg[n]=0;

	if(n<0){

		n=sendto(contfd, msg, strlen(msg), flags, (struct sockaddr *)&servAddr, sizeof(servAddr));

		alarm(1);
		n=recvfrom(contfd,msg,MAX_MSG,flags, (struct sockaddr *)&servAddr,&servLen);
		msg[n]=0;
	}

	printf("Server answered: %s\n",msg);

	sleep(5);	

	close(testfd);
	close(contfd);

#ifdef UNUSED1
	sprintf(msg, "test");
	n=sendto(testfd, msg, strlen(msg), flags, (struct sockaddr *)&servAddr, sizeof(servAddr));
	if(n<0){
		perror("sendto");
		exit(EXIT_FAILURE);		
	}
	int timeout = 0;
	while(1){
		/* Start timeout test */
		sprintf(msg, "test");
		n=sendto(testfd, msg, strlen(msg), flags, (struct sockaddr *)&servAddr, sizeof(servAddr));
		if(n<0){
			perror("sendto");
			break;
		}
		sleep(1);		

		printf("Sending ping %d\n",timeout);		
		sprintf(msg, "ping %d",timeout);
		n=sendto(contfd, msg, strlen(msg), flags, (struct sockaddr *)&servAddr, sizeof(servAddr));
		if(n<0){
			perror("sendto");
			break;
		}
		//servLen=sizeof(servAddr);
		alarm(1);
		n=recvfrom(testfd,recline,MAX_MSG,flags, (struct sockaddr *)&servAddr,&servLen);
		alarm(0);
		recline[n]=0;
		printf("debug %d %s, %d\n",n, recline,n);
		if(n>0){
			printf("Received %s from %s:%u\n",recline,inet_ntoa(servAddr.sin_addr),ntohs(servAddr.sin_port));
			timeout=timeout+20;
			
			sleep(timeout);
		}else{
			printf("timeout detected at %d seconds\n",timeout);
			break;	

		}
	
	
	
	
	}

	/* receive echoed message */
	echoLen = sizeof(echoAddr);
	while(1){
	    n = recvfrom(contfd, msg, MAX_MSG, flags,
		 (struct sockaddr *) &echoAddr, &echoLen);

	    if(n<0) {
		    printf("%s: cannot receive data \n",argv[0]);
		     continue;
		}

		 /* print received message */
		printf("%s: echo from %s:UDP%u : %s \n", 
		argv[0],inet_ntoa(echoServAddr.sin_addr),
		ntohs(echoServAddr.sin_port),msg);
	}

	}
#endif /* UNUSED1 */



	return EXIT_SUCCESS;

}






