/*
 *	Copyright (c) 2010, University of Helsinki
 *	
 *	All rights reserved.
 *	
 *	Redistribution and use in source and binary forms, with
 *	or without modification, are permitted provided that the
 *	following conditions are met:
 *	
 *		1. Redistributions of source code must retain the
 *		   above copyright notice, this list of conditions
 *		   and the following disclaimer.
 *	
 *		2. Redistributions in binary form must reproduce the
 *		   above copyright notice, this list of conditions and
 *		   the following disclaimer in the documentation and/or
 *		   other materials provided with the distribution.
 *	
 *		3. Neither the name of the University of Helsinki
 *		   nor the names of its contributors may be used
 *		   to endorse or promote products derived from this
 *		   software without specific prior written permission.
 *	
 *	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
 *	CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
 *	INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *	OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *	ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
 *	CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *	SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *	NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *	LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *	HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *	CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 *	OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 *	EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *	
 *	Contributors:
 *	
 *		Seppo Hätönen
 *		Aki Nyrhinen
 */

#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h> 
#include <signal.h>

#define LOCAL_SERVER_PORT 11000
#define MAX_MSG 1024
#define SERVER_TIMEOUT 3600

void sigalrm(int sig){
	        sig=sig;
}


int main(int argc, char *argv[]) {

	int fd, testfd, n, testLen, flags;
	socklen_t contLen;
	struct sockaddr_in testAddr, contAddr, servAddr;
	char msg[MAX_MSG];
	struct sigaction act;


	/* socket creation */
	fd=socket(AF_INET, SOCK_DGRAM, 0);
	if(fd<0) {
		printf("%s: cannot open socket \n",argv[0]);
		exit(1);
	}

	/* bind local server port */
	servAddr.sin_family = AF_INET;
	servAddr.sin_addr.s_addr = htonl(INADDR_ANY);
	
	if(argc==2){
		servAddr.sin_port = htons(atoi(argv[1]));
	}
	if(argc==3){
		servAddr.sin_addr.s_addr =inet_addr(argv[1]); 
		servAddr.sin_port= htons(atoi(argv[2]));
	}
	else
		servAddr.sin_port = htons(LOCAL_SERVER_PORT);
	//	servAddr.sin_port = 0;
	


	if(bind(fd, (struct sockaddr *) &servAddr,sizeof(servAddr))==-1){
		printf("%s: cannot bind port number %d \n", 
		   argv[0], LOCAL_SERVER_PORT);
		exit(1);
	}

	printf("%s: waiting for data on port UDP %u\n", 
	   argv[0],ntohs(servAddr.sin_port));


	flags = 0;
	struct sockaddr_in sa;
        socklen_t sa_len;

	sa_len = sizeof(sa);
	if(getsockname(fd, (struct sockaddr *)&sa,&sa_len)==-1){
		perror("getsockname()");
		return -1;
	}
	printf("Address is %s, Port is %u\n",inet_ntoa(sa.sin_addr), ntohs(sa.sin_port));


      /* set sighand so read() gets interrupted by SIGALRM */
	memset(&act, 0, sizeof act);
	act.sa_handler = sigalrm;
	sigfillset(&act.sa_mask);
	sigaction(SIGALRM, &act, NULL);


	int first_message = 0;
	/* server infinite loop */
	while(1) {

		/* init buffer */
		memset(msg,0,MAX_MSG);

		/* receive message */
		contLen = sizeof(contAddr);
		testLen = sizeof(testAddr);
		if(first_message==0)
			alarm(600);
		else
			alarm(SERVER_TIMEOUT);

		n = recvfrom(fd, msg, MAX_MSG, flags,
			 (struct sockaddr *) &contAddr, &contLen);

		if(n<0 && first_message == 0) { /* Check if first message was lost */
			printf("%s: No messages received, shutting down\n",argv[0]);
			close(fd);
			close(testfd);
			exit(EXIT_SUCCESS);
		}
		
		if(n<0){
			printf("No messages received for one hour, closing down\n");
			close(fd);
			close(testfd);
			exit(EXIT_SUCCESS);
		}
		
		first_message=1;

		msg[n]=0;
		alarm(0);

		/* print received message */
		printf("%s: Received message from %s:%u : %s \n", 
		   argv[0],inet_ntoa(contAddr.sin_addr),
		   ntohs(contAddr.sin_port),msg);
#ifdef UNUSED1
		if(strncmp(msg,"reg",3)==0){
			 /* socket creation */
			testfd=socket(AF_INET, SOCK_DGRAM, 0);
			if(testfd<0) {
				printf("%s: cannot open socket \n",argv[0]);
				sprintf(msg,"Register failed\n");
				sendto(fd,msg,strlen(msg),flags,(struct sockaddr *)&contAddr,contLen);
				continue;							        
			}


			/* bind local server port */
			testAddr.sin_family = AF_INET;
			//testAddr.sin_addr.s_addr = contAddr.sin_addr.s_addr;
                	testAddr.sin_addr.s_addr = htonl(INADDR_ANY);
                	testAddr.sin_port = 0;
	
			if(bind(testfd, (struct sockaddr *) &testAddr,sizeof(testAddr))==-1){
				printf("%s: cannot bind port number %d \n", 
		   			argv[0], ntohs(testAddr.sin_port));
				perror("bind");
				sprintf(msg,"Register failed\n");
				sendto(fd,msg,strlen(msg),flags,(struct sockaddr *)&contAddr,contLen);
				continue;

			}
			if(getsockname(testfd, (struct sockaddr *)&testAddr,&testLen)==-1){
	                	perror("getsockname()");
			                return -1;
			}
        		printf("Port is %u\n",ntohs(testAddr.sin_port));

			printf("Binded new UDP socket %s:%u\n", inet_ntoa(testAddr.sin_addr),ntohs(testAddr.sin_port));
			sprintf(msg,"Test port %u",ntohs(testAddr.sin_port));
		        sendto(fd,msg,strlen(msg),flags,(struct sockaddr *)&contAddr,contLen);

		}

#endif /* UNUSED1 */

		if(strncmp(msg,"ref",3)==0){
			printf("Received refresh from %s:%u msglen %d\n", inet_ntoa(contAddr.sin_addr), ntohs(contAddr.sin_port), strlen(msg));
                	sprintf(msg,"Refresh");
			sendto(fd,msg,strlen(msg),flags,(struct sockaddr *)&contAddr,contLen);
			sa=contAddr;
			sa_len=contLen;
		}

		if(strncmp(msg,"quit",4)==0){
                        printf("Received quit from %s:%u msglen %d\n", inet_ntoa(contAddr.sin_addr), ntohs(contAddr.sin_port), strlen(msg));
			printf("Closing down\n");
                	sprintf(msg,"bye");
			sendto(fd,msg,strlen(msg),flags,(struct sockaddr *)&contAddr,contLen);
			close(fd);
			close(testfd);
			exit(EXIT_SUCCESS);
		}


		if(strncmp(msg,"test",4)==0){
			printf("Received test from %s:%u msglen %d\n", inet_ntoa(contAddr.sin_addr), ntohs(contAddr.sin_port), strlen(msg));
			testAddr=contAddr;
			testLen=contLen;
		}

	
		if(strncmp(msg,"ping",4)==0){
			printf("Received ping from %s:%u msglen %d\n", inet_ntoa(contAddr.sin_addr), ntohs(contAddr.sin_port), strlen(msg));
			//printf("Viesti oli %s\n",msg);
			memset(msg,0,MAX_MSG);
			sprintf(msg,"ping");
			//printf("Debug %s asd\n",msg);
			int n = -10;
			n=sendto(fd,msg,strlen(msg),flags,(struct sockaddr *)&testAddr,testLen);
			//printf("Lähetettiin %d merkkiä\n",n);
			//n=sendto(fd,msg,strlen(msg),flags,(struct sockaddr *)&testAddr,testLen);
			//                        printf("Lähetettiin %d merkkiä\n",n);
		}

		if(strncmp(msg,"pong",4)==0){
			printf("Received pong from %s:%u msglen %d\n", inet_ntoa(testAddr.sin_addr), ntohs(testAddr.sin_port), strlen(msg));
			//printf("Viesti oli %s\n",msg);
			memset(msg,0,MAX_MSG);
			sprintf(msg,"pong");
			//printf("Debug %s asd\n",msg);
			int n = -10;
			n=sendto(fd,msg,strlen(msg),flags,(struct sockaddr *)&sa,sa_len);
			//printf("Lähetettiin %d merkkiä\n",n);
			//n=sendto(fd,msg,strlen(msg),flags,(struct sockaddr *)&testAddr,testLen);
			//                        printf("Lähetettiin %d merkkiä\n",n);
		}

		
		if(strncmp(msg,"ack",3)==0){
			printf("Received ack from %s:%u msglen %d\n", inet_ntoa(contAddr.sin_addr), ntohs(contAddr.sin_port), strlen(msg));
		}

		
		if(strncmp(msg,"block",5)==0){
			printf("Received block from %s:%u msglen %d\n", inet_ntoa(contAddr.sin_addr), ntohs(contAddr.sin_port), strlen(msg));

			
			memset(msg,0,MAX_MSG);
			printf("contAddr.sin_port %u testAddr.sin_port %u\n",ntohs(contAddr.sin_port),ntohs(testAddr.sin_port));
			if(ntohs(contAddr.sin_port)==ntohs(testAddr.sin_port)){
				printf("ON NE SAMAT!\n\n");
				printf("Block from %s:%u",inet_ntoa(contAddr.sin_addr),ntohs(contAddr.sin_port));
				sprintf(msg, "block from %s:%u",inet_ntoa(contAddr.sin_addr),ntohs(contAddr.sin_port));
			} else{
				printf("Binding changed!\n");
				sprintf(msg,"binding changed %s:%u",inet_ntoa(contAddr.sin_addr),ntohs(contAddr.sin_port));
			}

			int n = -10;
			printf("block status: %s\n",msg);
			n=sendto(fd,msg,strlen(msg),flags,(struct sockaddr *)&sa,sa_len);


			//printf("Lähetettiin %d merkkiä\n",n);
			//n=sendto(fd,msg,strlen(msg),flags,(struct sockaddr *)&testAddr,testLen);
			//                        printf("Lähetettiin %d merkkiä\n",n);
		}


/*
		int i;
		i=0;
		sleep(i);
		sprintf(msg,"Ping %d\n",i); 
		printf("Sending ping %d to %s:%u msglen %d\n",i, inet_ntoa(contAddr.sin_addr), ntohs(contAddr.sin_port), strlen(msg));
		sendto(fd,msg,strlen(msg),flags,(struct sockaddr *)&contAddr,contLen);
		i=i+5;	
*/
	}/* Server end */

	return 0;

}



